%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC/SUN4-SPEC.SL
% Title:          Some special things for SPARC PSL compiler
% Author:         Winfried Neun , ZIB Berlin
% Created:        9 Feb 1989
% Status:         Experimental
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Revisions:
% 12-Feb-1991  (Winfried Neun)
%   fixed bug in &alloctemp
% 22-August-1990 (Winfried Neun)
%   added code that attempts to compile setq to frame directly with the
%   frame element (which is a register in most cases) as &dest

% a special Linke to save the NOP instruction after final BA
(de &fillframeholes (code)
code)

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (cond ((FlagP FunctionName 'ForeignFunction)
              `((*ForeignLink ,FunctionName ,FunctionType ,NumberOfArguments)
            (*Exit 0)))
        ((get functionname 'exitopencode)
              `((*DeAlloc ,DeAllocCount)
                (*JCall ,FunctionName)))
        ((get functionname 'opencode)
              `((*Call ,FunctionName)
                (*exit ,DeAllocCount)))
        ((InternallyCallableP functionname)
              `((BA (InternalEntry ,functionname))
                (*dealloc ,Dealloccount)
                (nop)))
        (t    `((*dealloc ,Dealloccount)
          (*move  ($fluid symfnc) (reg t1))
          (*move (immediate (idloc ,functionname)) (reg t4))
          (sll (reg t4) 3 (reg t4))
          (*wplus2 (reg t1) (reg t4))
          (*move (displacement (reg t1) 0) (reg t4))
          (jmpl (displacement (reg t4) 0) (reg 0))
          (sethispec (immediate (idloc ,functionname)) (reg t3))))))

% change branch to exitlabel into *exit when framesize ne 0 too

(fluid '(exitt&))

(de &jumpopt (u)
    (when (equal (cadar u) exitt&) (rplaca u '(*exit (*framesize)))))

% turn off lapopt in faslend

(fluid '(*lapopt))

(setq *lapopt t)

(when (funboundp 'old-faslend) (copyd 'old-faslend 'faslend))

(de faslend ()
    (prog (*lapopt)
       (return (old-faslend))))

% open codes for cons ncons xcons


(put 'cons 'opencode
      '((*move ($fluid heaplast)      (reg 3))
        (*move ($fluid heaptrapbound) (reg 4))
        (sethi 16#120000 (reg 5))         % pair-tag
        (stx (reg 1) (indexed (reg 3) 0))  % Always sufficient memory for
	(sllx (reg 5) 32 (reg 5))
        (stx (reg 2) (indexed (reg 3) 8))  % 2 items left!
        (or (reg 3) (reg 5) (reg 1))      % the result
        (*wplus2 (reg 3) 16)
        (*move (reg 3) ($fluid heaplast))
        (*jumpwlessp (labelgen templabel) (reg 3) (reg 4))

        (*alloc 1)     % now that we are short of heap we have to prepare for gc
        (*move (reg 1) (frame 1))         % Reg 1 will be updated that way
        (*link !%reclaim expr 0)
        (*move (frame 1) (reg 1))         % Restore Reg 1; all other Regs
                                          % are not of interest
        (*dealloc 1)
        (labelref templabel))
)

(put 'ncons 'opencode
      '((*move ($fluid heaplast)      (reg 3))
        (*move ($fluid heaptrapbound) (reg 4))
        (sethi 16#120000 (reg 5))         % pair-tag
        (stx (reg 1) (indexed (reg 3) 0))  % Always sufficient memory for
	(sllx (reg 5) 32 (reg 5))
        (stx (reg g6)(indexed (reg 3) 8))  % 2 items left!
        (or (reg 3) (reg 5) (reg 1))      % the result
        (*wplus2 (reg 3) 16)
        (*move (reg 3) ($fluid heaplast))
        (*jumpwlessp (labelgen templabel) (reg 3) (reg 4))

        (*alloc 1)     % now that we are short of heap we have to prepare for gc
        (*move (reg 1) (frame 1))         % Reg 1 will be updated that way
        (*link !%reclaim expr 0)
        (*move (frame 1) (reg 1))         % Restore Reg 1; all other Regs
                                          % are not of interest
        (*dealloc 1)
        (labelref templabel))
)

(put 'xcons 'opencode
      '((*move ($fluid heaplast)      (reg 3))
        (*move ($fluid heaptrapbound) (reg 4))
        (sethi 16#120000 (reg 5))         % pair-tag
        (stx (reg 2) (indexed (reg 3) 0))  % Always sufficient memory for
	(sllx (reg 5) 32 (reg 5))
        (stx (reg 1) (indexed (reg 3) 8))  % 2 items left!
        (or (reg 3) (reg 5) (reg 1))      % the result
        (*wplus2 (reg 3) 16)
        (*move (reg 3) ($fluid heaplast))
        (*jumpwlessp (labelgen templabel) (reg 3) (reg 4))

        (*alloc 1)     % now that we are short of heap we have to prepare for gc
        (*move (reg 1) (frame 1))         % Reg 1 will be updated that way
        (*link !%reclaim expr 0)
        (*move (frame 1) (reg 1))         % Restore Reg 1; all other Regs
                                          % are not of interest
        (*dealloc 1)
        (labelref templabel)
)   )

(commentoutcode

% This code is intended for filling the 'holes' in the frame.
% e.g. in the following function
%
%  (de hugo (x) (cond ((null x) (call-another))
%                     (t (prog (a b c) ....))
%
% a frame of length 4 will be allocated, but it will not be preset (by nil)
% completely at entering the function but the frame elements for a b and c
% will be set (to nil) if the prog is entered.
% So if the (null x) case is taken we have 3 unset frame elements on the stack
% which contain random information and this is dangerous for gc.
% to resolve that problem, the following function will work on the input
% to pass-1-lap and insert nil settings where needed.

(fluid '(&fillframeholes &frame-numbers))

(de &fillframeholes (code)
   (prog (allo cmactoextern externfound ll lll nn linearcode restcode)
      (when (not &fillframeholes) (return code))
      (setq cmactoextern
                  '(*link *linke *call *jcall *cons *xcons *fastmkfltn))
             % trivial tests whether we have to work at all
      (unless (setq allo (findcmacro code '*alloc)) (return code))
      (setq ll cmactoextern)
      (while (and (not externfound) ll)
            (setq externfound (findcmacro code (car ll)))
            (setq ll (cdr ll)))
      (when (not externfound) (return code))

            % so we have to work, starting at the *alloc

      (setq nn (cadar allo))  %the frmae size
      (setq linearcode t)
      (setq restcode (cdr allo))

      (setq &frame-numbers nil)

      (while (and linearcode restcode)
         (setq linearcode (lookatinstruction (car restcode)))
         (setq restcode (cdr restcode)))

      (when linearcode (return code))

      (setq lll (setq ll (cons nil nil)))

             % search frame that have not been set, prepare code,
             % put into place after *alloc

      (for (from i 1 nn 1) (do
            (unless (memq i &frame-numbers)
                  (rplaca ll `(*move (reg nil) (frame ,i)))
%                 (prin2 "++ in ") (prin2 name&)
%                 (prin2 " presetting frame ")
%                 (prin2t i)
                  (push 'nil ll))))

      (when (eq lll ll) (return code))
      (rplacd lll (cdr allo))

      (rplacd allo (cdr ll))
      (return code)))

(de lookatinstruction (instr)
      (cond ((atom instr) t)
            ((and (eq (car instr) '*move)
                  (eqcar (caddr instr) 'frame))
                       (push (cadr (caddr instr)) &frame-numbers) 17)
            ((memq (car instr) '(*link *call *linke *jcall *jumpintype
                 *jumpeq *jump  *JumpType *JumpNotType *JumpNotInType
                 *jumpnoteq *jumpwlessp *jumpwgreaterp *jumpwleq *jumpwgeq
                 *jumpon)) nil)
            (t t)))


(de findcmacro (list pat)
     (cond ((atom list ) nil)
           ((eqcar (car list) pat) list)
           (t        (findcmacro (cdr list) pat ))))

(setq &fillframeholes t)

(compiletime (define-constant number-of-frame-regs 12))

(de &alloctemp (exp)
   (&alloctemp-aux exp t))

(de &alloctemp-aux (exp recu)   % from $pc/pass-2-3.sl 
  % Allocate a temporary REG for OPEN exprs.
  % Uses PREGS& to keep track
  % Used only when STATUS!&>1 to
  % set up destination.  Only special case is SETQ.  SETQ tries to put the
  % value of X:=... into a register containing X (keeps variables in the same
  % register if possible.

  % W. Neun, 25-May-1989:
  % For SUN4, 'small' frames are registers too, so tend to use the frame
  % position of the (setq result as dest instead of pulling a new tempreg

  (cond ((equal (car exp) 'setq)
         (cond ((equal (car (caddr exp)) 'setq)
                                       % Nested setq - move to actual RHS
                (&alloctemp-aux (caddr exp)
                                (and recu (eqcar (cadr (cadr exp)) '$LOCAL))))
               (t (prog (reg frm)
                        (cond % LHS variable already in reg?
                              ((and
                                (setq reg (&rassoc (cadr (cadr exp)) regs&))
                                (not (member (car reg) pregs&))
                               % and reg must be available
			 	(eqcar (car reg) 'reg))
				% must be a real reg
                               (return (car reg)))
                              ((and (eqcar (cadr (cadr exp)) '$LOCAL)
                                    recu
                                    (setq frm (assoc (cadr (cadr exp)) stomap&))
                                    (setq frm (cadr frm))
                                    (eqcar frm 'frame)
                                    (leq (cadr frm) number-of-frame-regs)
                                    (simpleenough (caddr exp)))
                               (return frm))
                              (t % Return the reg previously used for the var
                               (return (&tempreg))))))))
        (t % Just get a temp
           (&tempreg))))

(de simpleenough (exp)
   (prog (carexp)
          (cond ((atom exp) (return t)))
     loop (setq carexp (car exp))
          (cond ((atom carexp)
                   (when (memq carexp '(eq wgreaterp wlessp rplaca rplacd))
                               (return nil)))
                ((memq (car carexp) 
                      '(quote $fluid $global $local wconst car cdr quote)) nil)
                (t (return nil)))
          (setq exp (cdr exp))
          (when (pairp exp) (go loop))
          (return t)))
           
 
(de &dest (u)
  % Checks if U is the DEST& register
  (or (equal u dest&) 
      (and (eqcar u '$name) (&dest (cadr u)))
      (and (setq u (assoc u stomap&))
           (equal (cadr u) dest&))))

(de &storevar (u v)
  % The store generated by a SETQ
  (prog (vtype x)
        (&remrefs u)
        (cond ((equal (car u) '$local) % assumes $NAME
		(setq regs& (delasc (&getfrm u) regs&))
		(&storelocal u v))
              (t (&attach (list '*move v u))))
        (when (and (eqcar u 'frame) 
                       (leq (cadr u) number-of-frame-regs)) 
              (&remrefs (&getfvar u stomap&))
              (setq regs& (delasc u regs&))) 
        (when (or (&regp v)
                  (and (eqcar v 'frame)
                       (leq (cadr v) number-of-frame-regs)))
	  (setq regs& (&addrvals v regs& (&remregs u))))))
%%%       (setq regs& (&addrvals v regs& (list (&remregs u)))))))
 
(de &getfvar (v smap)
  %% From database in the form ((<a> <b> ... ) . . . ) retrieves the <a> from
  %% the first element whose <b> matches V in SMAP.  In the  (typical?) case
  %% where SMAP is STOMAP&, gets ($LOCAL <name>) from (FRAME <n>). /csp
  %% W. Neun: made error non fatal in (frame n) case.

  (cond ((null smap) (if (and (eqcar v 'frame)
		              (leq (cadr v) number-of-frame-regs))
			 v
			(&comperror (list "Compiler bug:" v "evaporated?"))))
        ((equal (cadar smap) v) (caar smap))
        (t (&getfvar v (cdr smap)))))

(de &LocalorRegnp (x) (or (&regnp x) (eqcar x '$Local)
			  (and (eqcar x 'frame)
				 (leq (cadr x) number-of-frame-regs))))
)
