%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:CONVEX-COMP.SL
% Description:  Compiler patterns for CONVEX PSL, plus a few CMACRO expanders
% Author:       G.Q. Maguire, M.L. Griss and Sam Sands
% Created:      16-Jan-1982
% Modified:     29-Oct-84 08:55:07 (Vicki O'Day)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%----------- Peep-hole optimization tables;
% interchage the two push operations in order top get correct sequence
(de &stopt (u)
  % OPTFN: Convert MOVEs + ALLOCS into PUSHES
  (cond ((and (equal (caadr u) '*alloc) (equal llngth& 1)
              (equal (cddar u) '((frame 1))))
         (rplacw u (cons (list '*push (cadar u)) (cddr u))))
        ((and (equal (caadr u) '*move) (equal (caaddr u) '*alloc)
              (equal llngth& 2) (equal (cddar u) '((frame 2)))
              (equal (cddadr u) '((frame 1))))
         (rplacw u
                 (cons (list '*push (cadar u))
                       (cons (list '*push (cadadr u)) (cdddr u)))))))


(PUT 'CARCDRPAT 'PATTERN                %  CAR, CDR
 '(nil nil
   ( ANY     ('*RETURN (FN A1)))))
   
% This one lets the cmacro make most of the decisions about the expansion
%  of its arguments. One case that is needed is if the first argument is
%  a reg. We do not want the one temp register allocated for memory to be
%  wasted on resolving a register, when the second arg might need it.

(PUT 'MEMORYPAT 'PATTERN      % MEMORY
 '(nil nil 
   ((Disp  Reg)     ('*RETURN (FN A2 A1)))
   ((Reg   Openfn)  ('*RETURN (FN A2 A1)))
   ((Any   Any)     ('*RETURN (FN A1 A2)))))


(PUT 'LOCPAT 'PATTERN '(NIL
  ('*SET DEST (FN A1))
  ((NOTIMM  DDEST)   (MAC TA2  A1)  ('*RETURN TA2))
  ( ANY              (MAC DEST A1))
  ))


(PUT 'ADDPAT 'PATTERN '(nil             %  WPLUS2               scs  
   ('*SET DEST (FN A1 A2)) 
   ((DEST     MOVEQ) ('*LOAD TD2 A2)  (MAC DEST TD2)) 
   ((MOVEQ    DEST)  ('*LOAD TD1 A1)  (MAC DEST TD1)) 
   ((USESDEST MOVEQ) ('*LOAD TD2 A2)  (MAC TD2 A1) ('*RETURN TD2)) 
   ((MOVEQ USESDEST) ('*LOAD TD1 A1)  (MAC TD1 A2) ('*RETURN TD1)) 
   ((ANY      MOVEQ) ('*LOAD DEST A2) (MAC DEST A1)) 
   ((MOVEQ    ANY)   ('*LOAD DEST A1) (MAC DEST A2)) 
   ((ANY      DEST)  (MAC A2 A1)) 
   ((REGNP    ANY)   (MAC A1 A2) ('*RETURN A1)) 
   ((ANY      REGNP) (MAC A2 A1) ('*RETURN A2)) 
   ((USESDEST 
           USESDEST) ('*LOAD T1   A1) (MAC T1 A2) ('*RETURN T1)) 
   ((ANY   USESDEST) ('*LOAD DEST A2) (MAC DEST A1)) 
   ((CONST ANY)      ('*LOAD DEST A2) (MAC DEST A1)) 
   ( ANY             ('*LOAD DEST A1) (MAC DEST A2))))
   

(PUT 'SUBPAT 'PATTERN '(nil             %  WDIFFERENCE           scs
   ('*SET DEST (FN A1 A2)) 
   ((REGNP MOVEQ)    ('*LOAD TD1 A2)  (MAC A1  TD1)   ('*RETURN A1)) 
   ((REGNP ANY)                       (MAC A1  A2)    ('*RETURN A1)) 
   ((DISPL USESDEST) ('*LOAD TA1 A1)  (MAC TA1 A2)    ('*RETURN TA1)) 
   ((ANY USESDEST)   ('*LOAD T1 A1)   (MAC T1 A2)     ('*RETURN T1)) 
   ((ANY MOVEQ)      ('*LOAD DEST A1) ('*USES DEST)   
                     ('*LOAD TD2  A2) (MAC DEST TD2)) 
   ( ANY             ('*LOAD DEST A1)                 (MAC DEST A2))))
   

(PUT 'UNARYPAT 'PATTERN '(nil           %  WMINUS, WNOT           scs
   ('*SET DEST (FN A1)) 
   ((SREGNP)         (MAC A1  A1) ('*RETURN A1)) 
   (ANY              (MAC TD1 A1) ('*RETURN TD1))))
   

                                    %  WAND, WOR, WXOR                scs
                                    %  TEMPORARY INEFFICIENT VERSION  scs
(PUT 'ANDORXORPAT 'PATTERN '(nil 
   ('*SET DEST (FN A1 A2)) 
   ((SREG   SREGDEST)                 (MAC A2 A1)) 
   ((SREGNP SREG)                     (MAC A1 A2)  ('*RETURN A1)) 
   ((SREG   SREGNP)                   (MAC A2 A1)  ('*RETURN A2)) 
   ((SREGNP ANY)      ('*LOAD TD2 A2) (MAC A1 TD2) ('*RETURN A1)) 
   ((ANY    SREGNP)   ('*LOAD TD1 A1) (MAC A2 TD1) ('*RETURN A2)) 
   ((ANY    ANY)      ('*LOAD TD1 A1) 
                      ('*LOAD TD2 A2) (MAC TD1 TD2) ('*RETURN TD1))))
   


(put 'setqpat 'pattern
     '(nil nil ((noval any notanyreg) ('!*store a2 a1))
           ((noval dest any) ('!*store a2 dest))
           ((noval usesdest any) ('!*load t1 a2) ('!*store t1 a1))
           ((noval any any) ('!*load dest a2) ('!*store dest a1))
           ((any dest) ('!*store dest a1)) ((dest any) ('!*store a2 dest))
           ((usesdest any) ('!*store a2 a1) ('!*store a2 dest))
           (any ('!*load dest a2) ('!*store dest a1))))

(put 'rplacpat 'pattern
     '(nil nil ((noval any any) ('!*store a2 (mac a1)))
           ((dest any) ('!*store a2 (mac a1)))
           ((usesdest any) ('!*store a2 (mac a1)) ('!*load dest a1))
           ((any dest) ('!*store a2 (mac a1)) ('!*load dest a1))
           ((any usesdest) ('!*store a2 (mac a1)) ('!*load dest a1))
           (any ('!*load dest a1) ('!*store a2 (mac dest)))))


(PUT 'SHIFTPAT 'PATTERN '(nil           %  WSHIFT, ASHIFT    scs
   NIL
   ((SREGNP CONST)       (MAC A1  A2) ('*RETURN A1)    ('*SET A1  (FN A1 A2))) 
   ((AREGNP LEQ2)        (MAC A1  A2) ('*RETURN A1)    ('*SET A1  (FN A1 A2))) 
   ((ANY    CONST)       ('*LOAD  TD1 A1) (MAC TD1 A2) ('*RETURN TD1) 
                                                       ('*SET TD1 (FN A1 A2))) 
   ((SREGNP SREGNP)      (MAC A1  A2)  ('*RETURN A1)   ('*SET A1  (FN A1 A2))
                         ('*DESTROY A2)) 
   ((ANY    SREGNP)      ('*LOAD  TD1 A1) 
                         (MAC TD1 A2)  ('*RETURN TD1)  ('*SET TD1 (FN A1 A2))
			 ('*DESTROY A2)) 
   ((SREGNP ANY)         ('*LOAD  TD2 A2) 
                         (MAC A1  TD2) ('*RETURN A1)   ('*SET A1  (FN A1 A2))
			 ('*DESTROY TD2)) 
   ((ANY    ANY)         ('*LOAD  TD1 A1) ('*LOAD TD2 A2) 
                         (MAC TD1 TD2) ('*RETURN TD1)  ('*SET TD1 (FN A1 A2))
			 ('*DESTROY TD2))
   ))
   

(PUT 'MKITEMPAT 'PATTERN '(nil          %   MKITEM   scs
   ('*SET DEST (FN A1 A2)) 
   ((SREGNP CONST)                         (MAC A1   A2) ('*RETURN A1)) 
   ((ANY    CONST DDEST)  ('*LOAD DEST A1) (MAC DEST A2)) 
   ((ANY    CONST)        ('*LOAD TD1  A1) (MAC TD1  A2) ('*RETURN TD1)) 
   ((REGNP  SREG)                          (MAC A1   A2) ('*RETURN A1)) 
   ((ANY    SREGDEST)     ('*LOAD TD1  A1) (MAC TD1  A2) ('*RETURN TD1)) 
   ((ANY    SREG  DDEST)  ('*LOAD DEST A1) (MAC DEST A2)) 
   ((REGNP  ANY)          ('*LOAD TD2  A2) (MAC A1   TD2) ('*RETURN A1)) 
   ((ANY    ANY   DDEST)  ('*LOAD DEST A1) ('*USES   DEST)
			  ('*LOAD TD2  A2) (MAC DEST TD2)) 
   ( ANY                  ('*LOAD TD1  A1) 
			  ('*LOAD TD2  A2) (MAC TD1  TD2) ('*RETURN TD1))))
   


                 


(put 'putfieldpat 'pattern
     '(nil nil ((noval any any any any) (mac a1 a2 a3 a4))
           (any (mac a1 a2 a3 a4) ('!*store a1 dest))))


(PUT 'FIELDPAT 'PATTERN '(nil           %  FIELD, SIGNEDFIELD   scs
   ('*SET DEST (FN A1 A2 A3)) 
   ((SREGNP ANY ANY)       (MAC A1   A1 A2 A3) ('*RETURN A1)) 
   ((ANY    ANY ANY DDEST) (MAC DEST A1 A2 A3)) 
   ( ANY                   (MAC TD1  A1 A2 A3) ('*RETURN TD1))))
   

(PUT 'ADDMEMPAT 'PATTERN '(nil nil      %  ADDMEM          scs
   
   ((NOVAL ANY MOVEQ)  ('*LOAD    TD2  A2) (MAC     A1   TD2)) 
   ((NOVAL ANY CONST)                      (MAC     A1   A2)) 
   ((NOVAL ANY SREG)                       (MAC     A1   A2)) 
   ((NOVAL ANY ANY)    ('*LOAD    TD2  A2) (MAC     A1   TD2)) 
   (      (ANY REGNP)  ('*WPLUS2  A2   LA1) ('*STORE A2   A1) ('*RETURN A2)) 
   ( (USESDEST ANY)    ('*LOAD    T2   A2) 
		       ('*WPLUS2  T2   LA1) ('*STORE T2   A1) ('*RETURN T2)) 
   ( ANY               ('*LOAD    DEST A2) 
		       ('*WPLUS2  DEST LA1) ('*STORE DEST A1))
   ))
   

(PUT 'SUBMEMPAT 'PATTERN '(nil nil      %  SUBMEM          scs
   
   ((NOVAL ANY MOVEQ)  ('*LOAD  TD2 A2)    (MAC     A1   TD2)) 
   ((NOVAL ANY CONST)                      (MAC     A1   A2)) 
   ((NOVAL ANY SREG)                       (MAC     A1   A2)) 
   ((NOVAL ANY ANY)    ('*LOAD    TD2  A2) (MAC     A1   TD2)) 
   ( (ANY  USESDEST)   ('*LOAD        T1   LA1) 
		       ('*WDIFFERENCE T1   A2)  ('*STORE T1   A1) ('*RETURN T1)) 
   (       ANY MOVEQ   ('*LOAD        DEST LA1) ('*USES DEST)
	               ('*LOAD        TD2  A2)
		       ('*WDIFFERENCE DEST TD2) ('*STORE DEST A1))
   (       ANY         ('*LOAD        DEST LA1) 
		       ('*WDIFFERENCE DEST A2)  ('*STORE DEST A1))
   ))
   

% PUT('MODMEMPAT1,'PATTERN,'(                      %  ???
%   NIL NIL
%    (ANY (MAC A1))));


(PUT 'GETUNSBYTE 'PATTERN '(nil 
  
  ('*SET DEST (FN A1)) 
  ((USESDEST) (MAC A1 TD2) ('*RETURN TD2)) 
  ((DDEST)    (MAC A1 DEST)) 
  ( ANY       (MAC A1 TD2) ('*RETURN TD2))))
  

(PUT 'PUTANYBYTE 'PATTERN '(nil nil 
  
  ((ANY SREG)                     (MAC A2  A1)) 
  ( ANY           ('*LOAD TD2 A2) (MAC TD2 A1))))
  

(PUT 'GETSGNHWORD 'PATTERN '(nil 
  
  ('*SET DEST (FN A1)) 
  ( ANY       (MAC A1 DEST))))
  

(PUT 'PUTANYHWORD 'PATTERN '(nil nil 
  
  
  ((ANY REG)                     (MAC A2 A1)) 
  ( ANY           ('*LOAD T2 A2) (MAC T2 A1))))
  

(PUT 'TVPAT 'PATTERN                    %  EQ, NE, WLEQ, etc FN
  '(&ALTDEST 
  ('*SET DEST (FN A1 A2)) 
% ((DEST   DEST)   SHOULD BE CAUGHT BY PASS 1
  ((AFRAME ZERO)   ('*USES DEST)       ('*LOAD TD1  A1) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 TD1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((ZERO   AFRAME) ('*USES DEST)       ('*LOAD TD2  A2) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 A1 TD2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    AREG)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((AREG   ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    ZERO)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((ZERO   ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((SMALL  ANY)    ('*USES DEST)       ('*LOAD TD1  A1) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 TD1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((ANY    SMALL)  ('*USES DEST)       ('*LOAD TD2  A2) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 A1 TD2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    SREG)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((SREG   ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    CONST)  ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((CONST  ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ( ANY            ('*USES DEST)       ('*LOAD T1    A1) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 T1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1))))
  


(PUT 'TVPAT1 'PATTERN '(nil             %  ATOM, PAIRP,  etc FN
    NIL
    ((SREGDEST) ('*LOAD T1   ''nil) (NEGMAC L1 A1 P2) 
                ('*LOAD T1   ''t)   ('*LBL  L1) ('*RETURN T1) 
		('*SET DEST (FN A1)) ('*DESTROY A1) ) 
    ((SREGNP)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 P2) 
                ('*LOAD DEST ''t)   ('*LBL  L1) 
		('*SET DEST (FN A1)) ('*DESTROY A1) ) 
    ( ANY       ('*USES DEST)       ('*LOAD TD1  A1) 
                ('*LOAD DEST ''nil) (NEGMAC L1 TD1 P2) 
                ('*LOAD DEST ''t)   ('*LBL  L1) 
		('*SET DEST (FN A1)) ('*DESTROY TD1) )
    ))
    


(PUT 'TSTPAT 'PATTERN '(nil             %  EQ, NE, WLEQ  etc   TST
    
    &FIXREGTEST 
    ((AFRAME ZERO) ('*LOAD TD1 A1) (MAC DEST TD1 A2)) 	%  TOO BAD   scs
    ((ZERO AFRAME) ('*LOAD TD2 A2) (MAC DEST A1 TD2)) 	%  TOO BAD   scs
    ((ANY    AREG)                 (MAC DEST A1  A2)) 
    ((AREG   ANY)                  (MAC DEST A1  A2)) 
    ((ANY    ZERO)                 (MAC DEST A1  A2)) 
    ((ZERO   ANY)                  (MAC DEST A1  A2)) 
    ((ANY  SMALL)  ('*LOAD TD2 A2) (MAC DEST A1 TD2)) 
    ((SMALL  ANY)  ('*LOAD TD1 A1) (MAC DEST TD1 A2)) 
    ((ANY    SREG)                 (MAC DEST A1  A2)) 
    ((SREG   ANY)                  (MAC DEST A1  A2)) 
    ((ANY  CONST)                  (MAC DEST A1  A2)) 
    ((CONST  ANY)                  (MAC DEST A1  A2)) 
    ( ANY          ('*LOAD T1 A1)  (MAC DEST T1  A2))))
    


(PUT 'TSTPAT2 'PATTERN '(nil            %  ATOM, PAIRP  etc  TST
     &SETREGS1 
     ((SREGNP)                  (MAC DEST  A1 P2) ('*DESTROY A1)) 
     ( ANY      ('*LOAD TD1 A1) (MAC DEST TD1 P2) ('*DESTROY TD1))))
     


(de &ALTAREG (ARGS) 
    (cond 
     (ARGS 
      % first A register   scs
      (cons (cond 
	     ((SREGP (CAR ARGS))
	      (prog (D I R ACTUALVAL)
		    (setq ACTUALVAL (&REMREGS (setq D (CAR ARGS))))
		    (setq I 9)
		    (RETURN (OR 
			     (WHILE (leq I MAXNARGS&)
				    (progn
				      (cond 
				       ((&SAMEVAL (setq R (&MKREG I))
						  D ACTUALVAL) 
					(RETURN R)))
				      (setq I (plus I 1))))
			     D))))
	      (t 
	       (CAR ARGS)))
	    (&ALTAREG (CDR ARGS))))))

(de &ALTDEST (ARGS) 
 (cond 
  ((OR (&USESDEST (CAR ARGS))                       % scs
       (&USESDEST (CADR ARGS)))
  (setq DEST& (&GETTEMP 'TD nil nil)))) 
 ARGS)




%%%%%%%%%%%%%%%%%%%% Versuch: 
 
(de &storevar (u v)
  % The store generated by a SETQ
  (prog (vtype x)
        (&remrefs u)
        (cond ((equal (car u) '$local) % assumes $NAME
               (&storelocal u v))
              (t (&attach (list '*move v u))))
        (when (&regp v)
   %%%%%  (setq regs& (&addrvals v regs& (list (&remregs u))))
          (setq regs& (&addrvals v regs& (&remregs u)))
        )))


 

