%----------------- Functions for general tests --------------------
(de &comtst (exp labl)
  % compiles boolean expression EXP.
  % If EXP has the same value as SWITCH!& then branch to LABL,
  % otherwise fall through;
  % REGS& are active registers for fall through,
  % REGS1& for branch;

  (prog (x fn reg)

    % First factor out NOT's to set up the SWITCH!&
    (while (and (eqcar exp 'eq) (equal (caddr exp) ''nil))
      (setq switch& (not switch&))
      (setq exp (cadr exp)))

    % SWITCH!& is assumed to be true by fn's with
    % a flip test
    (when (and (not switch&) (setq fn (get (car exp) 'fliptst)))
      (setq exp (cons fn (cdr exp))))

    % Dispatch a built in compiling function
    (if (setq fn (get (car exp) 'opentst))
      (if (atom fn)
	(apply fn (list exp labl))
	(&comopentst fn exp labl pregs&))
      (if (eqcar exp 'quote)
	(progn
	  (setq regs1& regs&)
	  (when (or (and switch& (cadr exp)) 
		    (and (not switch&) (not (cadr exp))))
	    (&attjmp labl))
	  )
	(&comtst (list 'ne exp ''nil) labl)))))

(de &comopentst (pat exp destlab pregs&)
  %? Pattern match entru for OPENTST functions
  (prog (oldstomap oldcode)
        (setq oldstomap stomap&)
        (setq oldcode codelist&)
        (&callopen pat destlab (&remopen (&tempreg) (&comlis (cdr exp)) 1) 
                   (car exp))
        (&cleartemps oldstomap oldcode)))

(de &remvarl (vars)
  % Remove variables to avoid name conflicts:  Hide variable names which match
  % new names when entering an inner function.  Other names will be available
  % as global info.  VARS is the list of new variable names, the result is a
  % list of protected stores, ie (var . slst-entry)
  %? What about REGS contents
  % Support for nested environments and branches; ie
  % name of VAR used
  (foreach x in vars collect (&protect x)))

(de &protect (u)
  % WON'T HANDLE MULTIPLE INSTANCES
  % Support for REMVARL
  % Delete all references to U from SLST!&
  % return the protected store, to be put back later
  (prog (x)
        (&remrefs u)
        %  scs
        (return (when (setq x (assoc u slst&))
                  (rplaca x nil)
                  (cons u x)))))

%  scs
(de &rstvarl (vars lst)
  % Restore a previous set of names.  VARS is the list of variables taken out
  % of the SLST. LST is the list of protected stores.  One or zero
  % stores for each variable. 
  %/ What of REGS
  (while vars
    (&rstvar (car vars) (car lst))
    (setq vars (cdr vars))
    (setq lst (cdr lst))))

(de &rstvar (var val)
  % Support for RSTVARL
  % Restore a particular variable and STORE
  (prog nil
        (&remrefs var)
        % Remove new occurance from REGS
        (&clrstr var)
        % Flush from SLST, removing unneeded Stores
        % in cas
        % Put back old deferred stores  SLSTS
        (&unprotect val)))

(de &unprotect (val)
  %restores VAL to SLST!&;
  (when val
    (rplaca (cdr val) (car val))))

%             scs
(de &storevar (u v)
  % The store generated by a SETQ
  (prog (vtype x)
        (&remrefs u)
        (cond ((equal (car u) '$local) % assumes $NAME    
               (&storelocal u v))
              (t (&attach (list '*move v u))))
        (when (&regp v)
	  (setq regs& (&addrvals v regs& (&remregs u))))))
%%%%%      (setq regs& (&addrvals v regs& (list (&remregs u)))))))

%  scs
(de &references (exp var)
  % True if expression EXP (probably ANYREG) references VAR.
  (or (equal exp var) (if (or (atom exp) (flagp (car exp) 'terminal))
        nil
        (&referencesl (cdr exp) var))))

(de &referencesl (exp var)
  % Support for REFERENCES
  (if (null exp)
    nil
    (or (&references (car exp) var) (&referencesl (cdr exp) var))))

(de &cfntype (fn)
  % Look up the type of a function using its CFNTYPE property if
  % any.  Return EXPR by default.
  (prog (x)
        (return (cond ((setq x (get fn 'cfntype)) (car x))
                      ((setq x (getd fn)) (car x))
                      (t 'expr)))))

(de &genlbl ()
  % General an internal name for a temp label
  (prog (l)
        (setq l (list 'label (&gensym)))
        (setq lblist& (cons (list l) lblist&))
        (return l)))

(de &getlbl (labl)
  % Look up the Internal name of label in a prog
  (prog (x)
        (setq x (assoc labl golist&))
        (when (null x)
          (&comperror (list "Compiler bug: missing label" labl)))
        (return (cdr x))))

(de &attlbl (lbl)
  % Attach a label to codelist.
  % Handle the (!*LBL x) (!*LBL y) case by making y alias for x
  (if (eq (caar codelist&) '*lbl)
    (&defeqlbl lbl (cadr (car codelist&)))
    (&attach (list '*lbl lbl))))

(de &attjmp (lbl)
  % Attach a jump to codelist.
  % handle (JUMP l) (LBL l) case
  (prog nil
        (when (eq (caar codelist&) '*lbl)
          (&defeqlbl lbl (cadr (car codelist&)))
          (&delmac codelist&))
        (when (&transferp codelist&)
          (return nil))
        (&attach (list '*jump lbl))))

(de &transferp (x)
  % Check if previously emited CMACRO is an unconditional
  % transfer (jump or call).
  (if (equal (caar x) '*noop)
    (&transferp (cdr x))
    (flagp (if (eq (caar x) '*link)
             (cadar x)
             (caar x)) 'transfer)))

(de &defeqlbl (lab1 lab2)
  % Make labels LAB1 and LAB2 aliases for each other
  (setq lblist& (&defeqlbl1 lblist& lab1 lab2)))

(de &defeqlbl1 (labs lab1 lab2)
  % Support for DEFEQLBL
  (cond ((member lab1 (car labs)) (if (member lab2 (car labs))
           labs
           (cons (append (&labclass lab2) (car labs)) 
                 (&delclass lab2 (cdr labs)))))
        ((member lab2 (car labs)) 
         (cons (append (&labclass lab1) (car labs)) 
               (&delclass lab1 (cdr labs))))
        (t (cons (car labs) (&defeqlbl1 (cdr labs) lab1 lab2)))))

(de &labclass (lab)
  %? Compute label equivalence classes
  (prog (temp)
        (setq temp lblist&)
        (while (and temp (not (member lab (car temp))))
          (setq temp (cdr temp)))
        (return (if temp
                  (car temp)
                  nil))))

(de &delclass (lab labs)
  % Remove a label from an equivalence class
  (if (member lab (car labs))
    (cdr labs)
    (cons (car labs) (&delclass lab (cdr labs)))))

(de &lbleq (lab1 lab2)
  % Check if LAB1eq to LAB2 (in its equivalence class)
  (member lab1 (&labclass lab2)))

(de &frame (u)
  % allocates space for U in frame;
  %? How about reuse of STOMAP entries?
  %? on exit from LAMBDA/PROG, mark as free?
  (prog (z res)
        (setq z (if (null stomap&)
                1
                (plus 1 (cadr (cadar stomap&)))))
        (setq res (&mkframe z))
        (setq stomap& (cons (list u res) stomap&))
        (return res)))

(de &getfrm (u)
  % GETFRM returns the frame location of a local variable
  (prog (x)
        (when (setq x (assoc u stomap&))
          (return (cadr x)))
        (&comperror (list "Compiler bug: lost variable" u))))

%*************************************************************************
% The following functions determine classes or properties of expressions *
%*************************************************************************
(de &anyreg (u)
  % !&ANYREG determines if U is an ANYREG expression
  %
  % ANYREG expressions are those expressions which may be loaded into any
  % register without the use of (visable) temporary registers.  It is assumed
  % that ANYREG expressions have no side effects.
  %
  % ANYREG expressions are defined as constants, variables, and ANYREG functions
  % whose arguments are ANYREG expressions.  Note that ANYREG functions are
  % not necessarily a part of ANYREG expressions; their arguments may not be
  % ANYREG expressions.
  (or (&constp u) (&varp u) (and (&anyregfnp u) (&anyregl (cdr u)))))

(de &anyregl (u)
  (or % Are all expression in list U Anyreg
      (null u) (and (&anyreg (car u)) (&anyregl (cdr u)))))

(de &anyregfnp (u)
  % !&ANYREGFNP is true when U is an ANYREG function.  The arguments are not
  % checked
  (&anyregp (car u)))

(de &openp (u)
  % does U open compile?
  (or (&constp u) (&varp u) 
      (and (or (&anyregfnp u) (&openfnp u)) (&openpl (cdr u)))))

(de &openpl (u)
  % Do each expression in list U opencompile
  (or (null u) (and (&openp (car u)) (&openpl (cdr u)))))

(de &openfnp (u)
  % true if U is an OPENfn
  (get (car u) 'openfn))

(de &constp (u)
  % True if U is a constant expression
  (and (idp (car u)) (flagp (car u) 'const)))

(de &varp (u)
  % True if U is a variable: (LOCAL x),(FLUID x), ...
  (and (pairp u) (flagp (car u) 'var)))

(de &regp (u)
  % True if U is a Register expressions (REG and EXTRAREG?)
  (and (pairp u) (flagp (car u) 'reg)))

(de &nosideeffectp (u)
  % True if the expression U has no side effects.  ANYREG expressions and
  % functions are assumed to have no side effects; other functions must be
  % flagged NOSIDEEFFECT.  All arguments to a function must also be NOSIDEEFFECT.
  (or (&anyreg u) 
      (and (or (&anyregfnp u) (flagp (car u) 'nosideeffect)) 
           (&nosideeffectpl (cdr u)))))

(de &nosideeffectpl (u)
  % Support for LIST of NOSIDEEFFECTS
  (or (null u) (and (&nosideeffectp (car u)) (&nosideeffectpl (cdr u)))))

%**********************************************************************
%  Basic register manipulation utilities
%**********************************************************************
(de &rval (r rgs)
  % Return the set of values in register R as determined by register list RGS
  (cond ((null rgs) nil)
        ((equal (caar rgs) r) (cdar rgs))
        (t (&rval r (cdr rgs)))))

(de &regval (r)
  % Normally, register contents are found in register list REGS!&.
  (&rval r regs&))

(de &addrvals (reg rgs vals)
  % Add values VALS to contents of REG in register list RGS
  (cond ((null rgs) (list (cons reg vals)))
        ((equal (caar rgs) reg) 
         (cons (cons (caar rgs) (append vals (cdar rgs))) (cdr rgs)))
        (t (cons (car rgs) (&addrvals reg (cdr rgs) vals)))))

(de &mkreg (num)
  %  Generates a tagged register from a register number
  (prog (aentry)
        (return %/ use a vector?
                (if (setq aentry 
                     (assoc num 
                      '((1 reg 1) (2 reg 2) (3 reg 3) (4 reg 4) (5 reg 5) 
                        (6 reg 6) (7 reg 7) (8 reg 8) (9 reg 9))))
                  (cdr aentry)
                  (list 'reg num)))))

(de &mkframe (num)
  % Used to generate a tagged FRAME loc from a frame number
  (prog (aentry)
        (return %/ use a vector?
                (if (setq aentry 
                     (assoc num 
                      '((1 frame 1) (2 frame 2) (3 frame 3) (4 frame 4) 
                        (5 frame 5) (6 frame 6) (7 frame 7) (8 frame 8) 
                        (9 frame 9))))
                  (cdr aentry)
                  (list 'frame num)))))

(de &rassoc (val rgs)
  % Find a register in register list RGS which contains VAL.  NIL is returned if
  % VAL is not present in RGS
  (cond ((null rgs) nil)
        ((member val (cdar rgs)) (car rgs))
        (t (&rassoc val (cdr rgs)))))

(de &repasc (reg val regl)
  % Replace the contants of REG in list REGL by the value VAL
  (cond ((null regl) (list (cons reg val)))
        ((equal reg (caar regl)) (cons (cons reg val) (cdr regl)))
        (t (cons (car regl) (&repasc reg val (cdr regl))))))

(de &rmerge (u)
  % RMERGE takes a list of register contents representing the information
  % present in the registers from a number of different ways to reach the same
  % place.  RMERGE returns whatever information is known to be in the registers
  % regardless of which path was taken.
  (if (null u)
    nil
    (prog (res contents)
          (setq res nil)
          (foreach rg in (car u) do (progn
                     (setq contents nil)
                     (foreach thing in (cdr rg) do (when (&inall thing 
                                                          (car rg) (cdr u))
                        (setq contents (cons thing contents))))
                     (when contents
                       (setq res (cons (cons (car rg) contents) res)))))
          (return res))))

(de &inall (thing rg lst)
  % Support for RMERGE, check in all branches
  (or (null lst) 
      (and (member thing (&rval rg (car lst))) (&inall thing rg (cdr lst)))))

(de &tempreg ()
  %  scs
  % Allocate  Compiler TEMP reg for open compilation
  % PREGS& protects registers currently in use
  (&temptypereg '&any))

(de &temptypereg (regkindp)
  %  scs 
  (prog (i r empty unprot)
        (setq empty (setq unprot nil))
        (setq i 1)
        (while (and (leq i maxnargs&) (not empty))
          (setq r (&mkreg i))
          (when (and (apply regkindp (list r)) (not (member r pregs&)))
            (cond ((and (leq i lastactualreg&) (null (&regval r))) 
                   (setq empty r))
                  ((not unprot) (setq unprot r))))
          (setq i (plus i 1)))
        (when empty
          (return empty))
        (when unprot
          (return unprot))
        (&comperror "Compiler bug: Not enough registers")
        (return '(reg error))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Mostly functions to get lists of expressions equivalent to some
%%% potential CMACRO argument. /csp
(de &remregs (u)
  %% Given one expression (suitable for CMACRO operand), returns list of
  %% alternative expressions equivalent to it.  Weird name.  I don't
  %% understand the final "else" case. /csp
  (cond ((&regp u) (&regval u))
        ((eqcar u 'frame) (list (&getfvar u stomap&)))
        ((or (&constp u) (&varp u)) (list u))
        (t (&remregsl u))))

(de &getfvar (v smap)
  %% From database in the form ((<a> <b> ... ) . . . ) retrieves the <a> from
  %% the first element whose <b> matches V in SMAP.  In the  (typical?) case
  %% where SMAP is STOMAP&, gets ($LOCAL <name>) from (FRAME <n>). /csp
  (cond ((null smap) (&comperror (list "Compiler bug:" v "evaporated?")))
        ((equal (cadar smap) v) (caar smap))
        (t (&getfvar v (cdr smap)))))

(de &remregsl (u)
  %% U is a CMACRO such as (*WPLUS2 (reg 1) (reg 2)).  Computes "ALLARGS" list
  %% from CDR (CMACRO args) and tacks op name onto each thing returned. /csp
  (foreach arg in (&allargs (cdr u)) collect (cons (car u) arg)))

(de &allargs (arglst)
  %% Uses &REMREGS to compute list of alternative, equivalent "expressions"
  %% for each arg, generates cross product of alternatives so that every
  %% element of &ALLARGS of (<a> <b> <c>)) has as its first element one of the
  %% alternatives for <a>, an alternative for <b>, and one for <c>. /csp
  (cond ((null arglst) nil)
        ((null (cdr arglst)) 
         (foreach val in (&remregs (car arglst)) collect (list val)))
        (t (&allargs1 (&remregs (car arglst)) (&allargs (cdr arglst))))))

(de &allargs1 (firstargs restargs)
  %% Support for ALLARGS /csp
  (prog (res)
        (setq res nil)
        (foreach a1 in firstargs do 
                 (foreach a2 in restargs do 
                  (setq res (cons (cons a1 a2) res))))
        (return res)))

%%% End of CMACRO alternative expression guys. /csp
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(de &remmrefs ()
  %? Remove MEMOERY references from values lists
  (setq regs& 
        (foreach r in regs& collect (cons (car r) (&remmrefs1 (cdr r))))))

(de &remmrefs1 (l)
  % Support for REMMREFS
  (cond ((null l) l)
        ((&refmemory (car l)) (&remmrefs1 (cdr l)))
        (t (cons (car l) (&remmrefs1 (cdr l))))))

(de &refmemory (exp)
  % Support for REMMREFS, check singel MEM cse
  (if (or (atom exp) (flagp (car exp) 'terminal))
    nil
    (or (member (car exp) '(memory car cdr)) (&refmemoryl (cdr exp)))))

(de &refmemoryl (l)
  % Support for REMEMORY
  (if (null l)
    nil
    (or (&refmemory (car l)) (&refmemoryl (cdr l)))))

(de &remvrefs ()
  %? Remove VARIABLE refs, and update SLST to ensure
  %? will be protected later
  (prog (s)
        (setq regs& 
              (foreach r in regs& collect 
                       (cons (car r) (&remvrefs1 (cdr r)))))
        % Slow version:
        %   SLST!& := FOR EACH S IN SLST!& CONC 
        %     IF !&EXTERNALVARP CAR S THEN NIL ELSE LIST S;
        % Faster version:
        %   while not null Slst!& and !&ExternalVarP car car Slst!& do
        %	Slst!& := cdr Slst!&;
        (setq s slst&)
        (while (not (null s))
          %              scs
          (progn
            (when (&externalvarp (car (car s)))
              (rplaca (car s) nil))
            (setq s (cdr s))))))

(de &remvrefs1 (l)
  %? support for REMVREFS
  (foreach thing in l conc (if (&refexternal thing)
             nil
             (list thing))))

(de &refexternal (exp)
  %? support for REMVREFS
  (cond ((atom exp) nil)
        ((&externalvarp exp) t)
        ((flagp (car exp) 'terminal) nil)
        (t (&refexternall (cdr exp)))))

(de &refexternall (exps)
  %? support for REMVREFS
  (if (null exps)
    nil
    (or (&externalvarp (car exps)) (&refexternall (cdr exps)))))

(de &externalvarp (u)
  %? support for REMVREFS
  (and (pairp u) (flagp (car u) 'extvar)))

(de &remrefs (v)
  % Remove all references to V from REGS!&
  (if (member (car v) '(memory car cdr))
    (&remmrefs)
    (setq regs& 
          (foreach r in regs& collect (cons (car r) (&remrefs1 v (cdr r)))))))

(de &remrefs1 (x lst)
  % Remove all expressions from LST which reference X
  (cond ((null lst) nil)
        ((&references (car lst) x) (&remrefs1 x (cdr lst)))
        (t (cons (car lst) (&remrefs1 x (cdr lst))))))

%************************************************************
%   Test functions
%************************************************************
(de &setbranch ()
  (setq slst& 
        (setq tail& 
              (setq branchpoint& 
                    (cons (list 'branch branchpoint& tail&) slst&))))
)

(de &nextbranch ()
  (when (and branchpoint& (ne slst& branchpoint&))
    (prog (s1 s2)
          (when slst&
            (setq s1 slst&)
            (while (and (setq s2 (cdr s1)) (ne s2 branchpoint&))
              (setq s1 s2))
            (rplacd s1 tail&)
            (setq tail& slst&))
          (unless s2
            (rplacd branchpoint& nil))
          (setq slst& branchpoint&)))
)

(de &unbranch ()
  (while branchpoint&
    (&nextbranch)
    (setq slst& tail&)
    (rplaca (car branchpoint&) nil)
    (setq tail& (caddr (car branchpoint&)))
    (setq branchpoint& (cadr (car branchpoint&)))))

(de &tstandor (exp labl)
  % Compile TEST expressions in COND, WHILE, etc as conditional Jumps
  (prog (flg flg1 fn lab2 regsl regs1l)
        %    scs
        %FLG is initial SWITCH!& condition;
        %FN is appropriate AND/OR case;
        %FLG1 determines appropriate switching state;
        (setq flg switch&)
        (setq switch& nil)
        (setq fn (eq (car exp) 'and))
        (setq flg1 (eq flg fn))
        (setq exp (cdr exp))
        (setq lab2 (&genlbl))
        (while exp
          (setq switch& nil)
          (if (and (null (cdr exp)) flg1)
            (progn
              (when fn
                (setq switch& t))
              (&comtst (car exp) labl)
              (setq regsl (cons regs& regsl))
              (setq regs1l (cons regs1& regs1l)))
            (progn
              (unless fn
                (setq switch& t))
              (if flg1
                (progn
                  (&comtst (car exp) lab2)
                  (setq regsl (cons regs1& regsl))
                  (setq regs1l (cons regs& regs1l)))
                (progn
                  (&comtst (car exp) labl)
                  (setq regsl (cons regs& regsl))
                  (setq regs1l (cons regs1& regs1l))))))
          (setq exp (cdr exp))
          (when exp
            (&setbranch))
          nil)
        %         scs
        (&attlbl lab2)
        (setq regs& (if (not flg1)
                (car regsl)
                (&rmerge regsl)))
        (setq regs1& (if flg1
                (car regs1l)
                (&rmerge regs1l)))
        (setq switch& flg)))

%************************************************************
%  Pass2 compile functions
%************************************************************

(de &special-comp-function (exp status)
  (let* ((fn-name  (first exp))
	(arg-count  (get fn-name 'argument-count))
	(compfn     (get fn-name 'compfn)))
    (if (and arg-count
	     (ne arg-count (length (rest exp))))
      (&comperror (list "Incorrect number of arguments to" fn-name "in" exp))
      (apply compfn (list exp status))
      )))

(de &comandor (exp status&) 
   % COMPFN: AND and OR chains
   (prog (fn labl regsl) 
      % Added source transformation to handle problem that (and) or
      % (or) didn't compile correctly.
      (cond ((equal exp '(and)) (setq exp '(and 't))) 
      ((equal exp '(or)) (setq exp '(or 'nil)))) 
      (setq fn (eq (car exp) 'and)) 
      (setq labl (&genlbl)) 
      (setq exp (cdr exp)) 
      (while exp 
      (progn (&comval (car exp) (cond ((cdr exp) 1) (t status&))) 
        %to allow for recursion on last entry;
	% BUG Busters strike again.  The following line will reset the
	% store optimization list to NIL.  This will fix the following
	% code:
        % (de foo () 
        %  (prog (x)
        %    (or (setf x 1) (setf x 2))
        %    (print x)))
	% Which was optimizing away the first setf.
	(setq slst& nil) 
        (setq regsl (cons regs& regsl)) 
	(cond ((cdr exp) (cond (fn (&jumpnil labl)) (t (&jumpt labl))))) 
	(setq exp (cdr exp)))) 
      (setq regs& (&rmerge regsl)) 
      (&attlbl labl)))

(de &comapply (exp status)
  (&comapply-aux exp status 'apply 'fastapply)
  )

(de &comapply-aux (exp status slow-function fast-function)

  % COMPFN:  Look for LIST in APPLY(foo, LIST(a,,,))
  %          or use FAST-APPLY or APPLY for special cases
  %
  % The new Cmacro *FAST-APPLY-LOAD is used to load the machine-specific 
  % register which hold the function being called.
  %
  % The test for number of arguments was changed from greaterp to geq to allow the
  % dreg model a register for the fastapply function argument.
  % It is safe but unnecessary for the previous implementations. /csp

  (let* ((fn     (second exp))
	 (args   (consed-list (third exp))))
    (if (or (eq args 'not-a-consed-list)
	    (geq (length args) maxnargs&))
      (&call slow-function (rest exp) status)
      (let ((n-args (length args)))
	(setf args (append args (list fn)))
	(&comload args)
	(&attach (list '*fast-apply-load (&mkreg (plus n-args 1))))
	(&attach (list '*link fast-function 'expr n-args))
	(setf regs& nil)
	(&remvrefs)
	))))

(de consed-list (exp)
  % If exp is a list that can be decomposed by the compiler, a list
  % of its components is returned.  If not, 'NOT-A-CONSED-LIST is returned.
  (cond
   ((atom exp)
    'not-a-consed-list
    )
   ((memq (first exp) '(list ncons list1 list2 list3 list4 list5))
    (rest exp)
    )
 % ((eq (first exp) 'quote)  wrong handling modified; HM, 21-dec-89
 %  (second exp)
 %  )
   ((and (eqcar exp 'quote) (pairp (second exp)))
    (mapcar (second exp) (function mkquote))
    )
   (t
    'not-a-consed-list
    )
   ))

(de &comcodeapply (exp status)
  (&comapply-aux exp status 'codeapply 'fast-codeapply)
  )

(de &comcond (exp status&)

  % COMPFN: compiles conditional expressions;
  %         registers REGS!& are set for dropping through,
  %         REGS1  are set for a branch;
  %
  % EXP must be of the form: 
  %    (COND ((test1 form1)
  %           ...
  %           (testn formn) 
  %           ((QUOTE T) (formn+1)))
  %
  % NOTE: The code generated contains an extraneous jump from the end of 
  %       formn+1 to the bottom of the cond (the next instruction).  It
  %       is assumed that the optimizer will take care of this.

  (let* ((clauses       (rest exp))
	 (bottom-label  (&genlbl))
	 branch-regs              % Register values from each jump
                                  % to bottom-label.
	 branchpoint&
	 tail&
	 )

%print (list "branchpoint&" branchpoint& slst&))
    (unless (and clauses
		 (for (in x clauses) (always (equal (length x) 2)))
		 (equal (first (lastcar clauses)) '(quote t)))
      (&comperror (bldmsg "Bad COND in pass 2 of compiler: %r" exp))
      )
    
%print (list "branchpoint&" branchpoint& slst&))
    (for (in clause clauses)
	 (do
	  (let ((next-case-label (&genlbl))
		(switch&         nil)        % Used by &COMTST
		regs1&                       % Set by &COMTST
		)
	    (&comtst (first clause) next-case-label)
	    (&setbranch)
	    (&comval (second clause) status&)
	    (&nextbranch)
	    (&attjmp bottom-label)
	    (setq branch-regs (cons regs& branch-regs))
	    (setq regs& regs1&) % restore register status for next iteration;
	    (&attlbl next-case-label)
	    )))
    (setq regs& (&rmerge (cons regs& branch-regs)))
    (&attlbl bottom-label)
    (&unbranch)
    ))

(de &comcons (exp status&)
  % COMPFN: for CONS, in favor of NCONS and XCONS special cases
  (setf exp (rest exp))
  (cond ((equal (cadr exp) ''nil) (&call 'ncons (list (car exp)) status&))
        ((and (member (cadr exp) (&regval '(reg 1))) (&openp (car exp))) 
         (&call 'xcons (reverse exp) status&)) %  scs
        (t (&call 'cons exp status&))))   %  scs

(de &comcase (exp status&)
  % COMPFN: Compile CASE statement
  %         Look for special cases as JUMPON, JUMPEQ , JUMPWITHIN etc
  (prog (bottomlab regs1& jumps exps elselab high low saveregs jmps jlist 
                   ranges table branchpoint& tail&)
        %  scs
        (setq bottomlab (&genlbl))
        (setq regs1& nil)
        (&comval (cadr exp) 1)
        (setq jumps (setq exps nil))
        (foreach thing on (cddr exp) do 
                 (prog (lab)
                       (setq lab (&genlbl))
                       (setq jumps 
                        (nconc jumps (list (list (caar thing) lab))))
                       (setq exps 
                        (nconc exps (list (list lab (cadar thing)))))
                       (when (null (cdr thing))
                         (if (not (null (caar thing)))
                           (if (greaterp status& 1)
                             (progn
                               (setq regs1& (cons regs& regs1&))
                               (setq elselab bottomlab))
                             (setq exps 
                              (nconc exps 
                               (list (list (setq elselab (&genlbl)) ''nil)))))
                           (setq elselab lab)))))
        (setq ranges nil)
        (setq table nil)
        (foreach jmp in jumps do 
                 (foreach num in (car jmp) do (if (eqcar num 'range)
                    (prog (high low)
                          (setq low (&getnum (cadr num)))
                          (setq high (&getnum (caddr num)))
                          (when (geq high low)
                            (if (lessp (difference high low) 6)
                              (for (from i low high 1) 
                               (do 
                                (setq table (&instbl table i (cadr jmp)))))
                              (setq ranges 
                               (nconc ranges 
                                (list (list low high (cadr jmp))))))))
                    (setq table (&instbl table (&getnum num) (cadr jmp))))))
        (foreach r in ranges do 
                 (&attach (list '*jumpwithin (caddr r) (car r) (cadr r))))
        (while table
          (setq jmps (list (car table)))
          (setq low (setq high (caar table)))
          (setq jlist (list (cadar table)))
          (while (and (cdr table) (lessp (car (cadr table)) (plus high 5)))
            (setq table (cdr table))
            (while (lessp high (difference (caar table) 1))
              (setq high (plus high 1))
              (setq jlist (nconc jlist (list elselab))))
            (setq high (plus high 1))
            (setq jlist (nconc jlist (list (cadar table))))
            (setq jmps (nconc jmps (list (car table)))))
          (if (lessp (length jmps) 4)
            (foreach j in jmps do 
                     (&attach 
                      (list '*jumpeq (cadr j) '(reg 1) 
                       (list 'wconst (car j)))))
            (prog (reg)
                  (if (cdr table)
                    (progn
                      (setq reg '(reg 2))
                      %  scs
                      (&lreg reg '(reg 1)))
                    (setq reg '(reg 1)))
                  (&attach 
                   (cons '*jumpon (cons reg (cons low (cons high jlist)))))
                  (setq regs& (delasc reg regs&))))
          %  scs
          (setq table (cdr table)))
        (&attjmp elselab)
        (setq saveregs regs&)
        (&setbranch)
        %      scs
        (foreach thing in exps do (progn
                   (&attlbl (car thing))
                   (setq regs& saveregs)
                   (when (cadr thing)
                     (&comval (cadr thing) status&))
                   (&nextbranch)
                   %      scs
                   (unless (&transferp codelist&)
                     (&attjmp bottomlab)
                     (setq regs1& (cons regs& regs1&)))))
        (&attlbl bottomlab)
        (setq regs& (&rmerge regs1&))
        (&unbranch)))

%      scs
(de &instbl (tbl i l)
  % Support for COMCASE, one branch
  (cond ((null tbl) (list (list i l)))
        ((lessp i (caar tbl)) (cons (list i l) tbl))
        ((equal i (caar tbl)) (&comperror (list "Ambiguous case" tbl)))
        (t (cons (car tbl) (&instbl (cdr tbl) i l)))))

(de &getnum (x)
  % Support for CASE, check numeric tag
  (if (integer-constant? x)
    (constant-value x)
    (&comperror (list "Number expected for CASE label" x))
    ))

(de &comgo (exp status&)
  % COMPFN: Compile GO inside PROG.
  %/ Should SLST& be cleared like this?
  (progn
    (cond ((greaterp status& 2) (&attjmp (&getlbl (cadr exp))) 
           (setq slst& nil))
          (t % scs
             (&comperror (list exp "invalid go"))))))

(de &comidapply (exp status)
  (&comapply-aux exp status 'idapply 'fast-idapply)
  )

(de &comprog (exp status&)
  % COMPFN: for PROG compiles program blocks;
  %/ need to improve SLST handling, binding
  %/ see LAMBDA
  (prog (oldstomap oldcode i alsts& golist& pg proglis exitt& exitregs&)
        %  scs
        (setq exitregs& nil)
        (setq proglis (cadr exp))
        (setq exp (cddr exp))
        (setq exitt& (&genlbl))
        (setq pg (&remvarl proglis))
        %protect prog variables;
        (setq oldcode codelist&)
        %  scs
        (setq oldstomap stomap&)
        %  scs
        (setq alsts& (&varbind proglis nil))
        (foreach x in exp do (when (atom x)
                   (setq golist& (cons (cons x (&genlbl)) golist&))))
        (while exp
          (if (atom (car exp))
            (progn
              (&attlbl (&getlbl (car exp)))
              (setq regs& nil))
            (&comval (car exp) (if (greaterp status& 1)
                       4
                       3)))
          %   scs
          (setq exp (cdr exp)))
        (when (and (not (&transferp codelist&)) (lessp status& 2))
          (&lreg1 ''nil))
        (&attlbl exitt&)
        (setq regs& (&rmerge (cons regs& exitregs&)))
        (&freerstr alsts& status&)
        (&fixfrm oldstomap oldcode)
        %  let's go for it  scs
        (&rstvarl proglis pg)))

(de &comprogn (exp status&)
  % COMPFN: Compile PROGN cases
  (prog nil
        (setq exp (cdr exp))
        (when (null exp)
          (return (&comval ''nil status&)))
        (while (cdr exp)
          (&comval (car exp) (if (lessp status& 2)
                     2
                     status&))
          (setq exp (cdr exp)))
        (&comval (car exp) status&)))

(de &comreturn (exp status&)
  % COMPFN: Compile RETURN inside PROG
  %/ Should be better SLSTS handling, last return
  %/ Cf LAMBDA case
  (setq exp (cdr exp))
  %  test for 1,2?   scs
  (when (or (lessp status& 4) (not (&nosideeffectp (car exp))))
    (&comload exp))
  (setq slst& nil)
  (setq exitregs& (cons regs& exitregs&))
  (&attjmp exitt&))

(de &delmac (x)
  % Delete macro CAR X from CODELIST!&
  (rplaca x '(*noop)))

%*************************************************************
%              Pass 3
%*************************************************************
%----------------------------- Post Code Generation Fixups; 
(de &pass3 ()
  % Pass 3 - optimization.
  %    The optimizations currently performed are:
  % 1. Deletion of stores not yet picked up from SLST!&.
  % 2. Removal of unreachable macros.
  % 3. A peep hole optimizer, currently only optmizing LBL macros.
  % 4. Removal of common code chains
  % 5. Changing LINK to LINKE where possible
  % 6. Squeezing out unused frame locations and mapping the stack onto
  %    the registers.
  % Other functions of PASS3 are to tack exit code on the end and reverse
  % the code list.
  (progn
    (foreach j in slst& do (when (car j)
               (&delmac (cdr j))))
    %    scs
    (&attlbl exitt&)
    (&attach '(*exit (*framesize)))
    (&remcode t)
    (&fixlabs)
    (&fixchains)
    (&fixlinks)
    (&remcode nil)
    (&fixfrm nil nil)
    %   scs
    (&peepholeopt)
    (&remcode nil)
    (setq codelist& (reversip codelist&))
    nil))

(de &insertmac (place mac)
  % Insert a CMACRO into CODELIST
  (rplacw place (cons mac (cons (car place) (cdr place)))))

(de &deletemac (place)
  % Delete a CMACRO from CODELIST
  (rplacw place (cdr place)))

(de &remcode (keeptop)
  %? Remove a block of code  for Unused labels
  (prog (unusedlbls)
        (setq unusedlbls (&unusedlbls keeptop))
        (&remunusedmac unusedlbls)
        (while (setq unusedlbls (&unusedlbls keeptop))
          (&remunusedmac unusedlbls))))

(de &unusedlbls (keeptop)
  %? Which labels are not used?
  (prog (used unused)
        (setq used nil)
        (setq unused lblist&)
        (when keeptop
          (setq used (cons (&labclass toplab&) used))
          (setq unused (&delclass toplab& unused)))
        (foreach mac in codelist& do (when (neq (car mac) '*lbl)
                   (foreach fld in (cdr mac) do (when (and 
                                                       (eqcar fld 'label) 
                                                       (&classmember fld 
                                                        unused))
                      (setq used (cons (&labclass fld) used))
                      (setq unused (&delclass fld unused))))))
        (setq lblist& used)
        (return unused)))

(de &classmember (lab classes)
  % support for label hacking
  (if (null classes)
    nil
    (or (member lab (car classes)) (&classmember lab (cdr classes)))))

(de &remunusedmac (unusedlabs)
  % Remove the !*NOOPs, or LBLS after transfers
  (prog (p q r)
        (setq codelist& (setq p (reversip codelist&)))
        (while (cdr p)
          (setq q (cdr p))
          (if (or (equal (caar q) '*noop) 
                  (and (&transferp p) (neq (caar q) '*lbl)) 
                  (and (equal (caar q) '*lbl) 
                       (&classmember (cadar q) unusedlabs)))
            (rplacd p (cdr q))
            (setq p (cdr p))))
        (setq codelist& (reversip codelist&))))

(de &fixlinks ()
  %
  % replace LINK by LINKE where appropriate
  %
  (when (and (not *nolinke) (not freebound&))
    (prog (switched)
          (foreach inst on codelist& do 
                   (prog (saverest)
                         (when (or 
                                (and exitt& 
                                 (equal (first (first inst)) '*jump) 
                                 (equal (second (first inst)) exitt&))
                                (equal (first (first inst)) '*exit))
                           (when (equal (first (second inst)) '*lbl)
                             (when (equal (first (third inst)) '*link)
                               (setq inst (cdr inst))
                               (setq saverest t)))
                           (when (equal (first (second inst)) '*link)
                             (if (and (eq (second (second inst)) name&) 
                                  *r2i)
                               (rplaca (rest inst) (list '*jump toplab&))
                               (rplaca (rest inst) 
                                (cons '*linke 
                                 (cons '(*framesize) (rest (second inst))))))
                             (unless saverest
                               (&deletemac inst)))))))))

(de &peepholeopt ()
  % table driven 'peep-hole' optimization for various cases;
  (prog (x z)
        (setq z codelist&)
        (while z
          (cond ((equal (caar z) '*noop) (&deletemac z))
                ((or (not (setq x (get (caar z) 'optfn))) 
                     (not (apply x (list z))))
                 (setq z (cdr z)))))))

%----------- Peep-hole optimization tables; 
(de &stopt (u)
  % OPTFN: Convert MOVEs + ALLOCS into PUSHES
  (cond ((and (equal (caadr u) '*alloc) (equal llngth& 1) 
              (equal (cddar u) '((frame 1))))
         (rplacw u (cons (list '*push (cadar u)) (cddr u))))
        ((and (equal (caadr u) '*move) (equal (caaddr u) '*alloc) 
              (equal llngth& 2) (equal (cddar u) '((frame 2))) 
              (equal (cddadr u) '((frame 1))))
         (rplacw u 
                 (cons (list '*push (cadadr u)) 
                       (cons (list '*push (cadar u)) (cdddr u)))))))

(de &lblopt (u)
  %? OPTFN: Convert some LABLE/JUMP chains
  (prog (z)
        (when (equal (cadr u) '*lbl)
          (&defeqlbl (cadr u) (cadr (cdr u)))
          (rplacd u (cddr u))
          (return t))
        (cond ((and (cdadr u) (eqcar (cadadr u) 'label) 
                    (&lbleq (cadar u) (cadadr u)))
               (return (rplacw (cdr u) (cddr u))))
              ((and (equal (caadr u) '*jump) 
                    (setq z (get (caaddr u) 'negjmp)) 
                    (&lbleq (cadar u) (cadr (caddr u))))
               (return (progn
                         (setq z 
                          (cons z (cons (cadadr u) (cddr (caddr u)))))
                         (rplacd u (cons z (cdddr u)))
                         t)))
              (t (return nil)))))

(de &jumpopt (u)
  % OPTFN: Convert JUMP to exit label into an EXIT
  (when (and (equal (cadar u) exitt&) (equal llngth& 0))
    (rplaca u '(*exit (*framesize)))))

(de &fixchains ()
  %? Fix common code chains
  (prog (lab)
        (foreach labcode on codelist& do 
                 (cond ((equal (caar labcode) '*lbl) % OR CAAR LABCODE = '!*JUMP	% croaks on this one
                        (progn
                          (setq lab (cadar labcode))
                          (foreach jumpcode on (cdr labcode) do (when (and 
                                                                       (equal
                                                                        (caar
                                                                         jumpcode)
                                                                        '*jump)
                                                                       (equal
                                                                        (cadar
                                                                         jumpcode)
                                                                        lab))
                             (&movejump labcode jumpcode)))))))))

(de &movejump (labcode jumpcode)
  %? OPTFN: Interchanges Moves and jumps inside Loops
  (when (equal (cadr labcode) (cadr jumpcode))
    (prog (lab)
          (repeat (progn
                    (when (equal (cadr labcode) (cadr jumpcode))
                      (setq jumpcode (cdr jumpcode))
                      (setq labcode (cdr labcode)))
                    (while (equal (caadr labcode) '*lbl)
                      (setq labcode (cdr labcode)))
                    (while (equal (caadr jumpcode) '*lbl)
                      (setq jumpcode (cdr jumpcode)))
                    nil)
                  (not (equal (cadr jumpcode) (cadr labcode))))
          (if (equal (caar labcode) '*lbl)
            (rplacd jumpcode 
                    (cons (list '*jump (cadr (car labcode))) 
                     (cdr jumpcode)))
            (progn
              (setq lab (&genlbl))
              (rplacd jumpcode (cons (list '*jump lab) (cdr jumpcode)))
              (rplacd labcode (cons (list '*lbl lab) (cdr labcode))))))))

(de &fixfrm (oldstomap oldcode)
  %       scs
  % Squeeze out HOLES
  %/ Should change FIXFRM to do sliding squeeze, not reorder;
  (prog (n freeregs framesused r fr p holemap)
        (setq framesused (&getframes codelist& oldcode nil))
        (setq n 0)
        (foreach var in oldstomap do (progn
                   (setq n (max n (cadr (setq fr (cadr var)))))
                   (setq framesused (delete fr framesused))))
        (setq llngth& n)
        (unless framesused
          (return nil))
        (foreach frame in framesused do 
                 (setq llngth& (max llngth& (cadr frame))))
        (setq freeregs (unless (and freebound& (not *useregfluid))
                (&freeregs codelist& oldcode nil)))
        (setq holemap nil)
        (foreach frame in framesused do 
                 (setq holemap 
                       (cons 
                        (list frame (if freeregs
                           (progn
                             (setq r (car freeregs))
                             (setq freeregs (cdr freeregs))
                             (setq regs& (delasc r regs&))
                             r)
                           (&mkframe (setq n (plus n 1)))))
                        holemap)))
        (when (equal n llngth&)
          (return nil))
        (setq llngth& n)
        (setq p codelist&)
        (while (not (eq p oldcode))
          (rplaca p (&macrosubst (car p) holemap))
          (setq p (cdr p)))))

(de &getframes (code oldcode res)
  %? Support FIXFRAME?
  (if (eq code oldcode)
    res
    (&getframes (cdr code) oldcode (&getframes1 (cdar code) res))))

(de &getframes1 (macargs res)
  %? Support FIXFRAME
  (if (null macargs)
    res
    (&getframes1 (cdr macargs) (&getframes2 (car macargs) res))))

(de &getframes2 (macarg res)
  %? Support FIXFRAME
  (cond ((or (atom macarg) (&varp macarg) (&constp macarg) (&regp macarg)) 
         res)
        ((eqcar macarg 'frame) (if (member macarg res)
           res
           (cons macarg res)))
        (t (&getframes1 (cdr macarg) res))))

(de &freeregs (start stop exitflag)
  %       scs
  (prog (i r usedregs freeregs)
        (when (equal (setq usedregs (&regsused start stop nil exitflag)) 
                     'all)
          (return 'nil))
        (for (from i lastactualreg& 1 -1) 
             (do (unless (member (setq r (&mkreg i)) usedregs)
                   (setq freeregs (cons r freeregs)))))
        (return freeregs)))

(de &regsused (start stop regsused exitflag)
  %        scs
  % Find which registers are used.  'ALL is returned if all are used.
  % Supports INSTALLDESTORY and Reg mapping
  (if (eq start stop)
    regsused
    (prog (mac instr fn nargs)
          (setq mac (car start))
          (setq instr (car mac))
          (when (and (setq nargs 
                      (cond ((eq instr '*link) (setq fn (cadr mac)) 
                             (cadddr mac))
                            ((or (eq instr '*linke) (eq instr '*linkef)) 
                             (setq fn (caddr mac)) (car (cddddr mac)))))
                     (not exitflag))
            (while (greaterp nargs 0)
              (setq regsused (adjoin (&mkreg nargs) regsused))
              (setq nargs (sub1 nargs))))
          (return (cond ((or (equal instr '*link) 
                          (and (equal instr '*linke) exitflag))
                         (cond ((equal fn name&) (if exitflag
                                  (&regsused (cdr start) stop regsused 
                                   exitflag)
                                  'all))
                               ((and *usingdestroy 
                                 (setq dest& (get fn 'destroys)))
                                (setq regsused (union dest& regsused)) 
                                (&regsused (cdr start) stop regsused 
                                 exitflag))
                               (t 'all)))
                        ((or (equal instr '*linkf) 
                          (and (equal instr '*linkef) exitflag))
                         'all)
                        (t 
                         (&regsused (cdr start) stop 
                          (&regsused1 regsused (cdr mac)) exitflag)))))))

(de &regsused1 (regsused args)
  %   scs
  %? Support REGSUSED
  (prog nil
        (foreach a in args do (setq regsused (&regsused2 regsused a)))
        (return regsused)))

(de &regsused2 (regsused arg)
  %     scs
  % Support REGSUSED
  (cond ((atom arg) regsused)
        ((not (atom (car arg))) (&regsused1 regsused arg))
        ((&constp arg) regsused)
        ((and (equal (car arg) 'reg) (numberp (cadr arg))) 
         (adjoin arg regsused))
        (t (&regsused1 regsused (cdr arg)))))

(de &reformmacros ()
  % Converts some pseudo-s into numbers
  (prog (finaltransform)
        (setq finaltransform (list (list '(*framesize) llngth&)))
        (foreach mac on codelist& do 
                 (rplaca mac (&macrosubst (car mac) finaltransform)))))

(de &fixlabs ()
  %  scs
  %? Fixup LABELS and JUMPS ?
  (prog (transform u instr0 instr long)
        (setq transform nil)
        (foreach lab in lblist& do 
                 (foreach eqlab in (cdr lab) do 
                  (setq transform (cons (list eqlab (car lab)) transform))))
        (foreach mac on codelist& do (progn
                   (setq instr0 (car mac))
                   (when (neq instr0 
                          (setq instr (&macrosubst instr0 transform)))
                     (when (setq long (get (car instr) 'longbranch))
                       (rplaca instr long))
                     (rplaca mac instr))))
        (when (setq u (assoc exitt& transform))
          (setq exitt& (cadr u)))
        (when (setq u (assoc toplab& transform))
          (setq toplab& (cadr u)))
        (setq lblist& (foreach lab in lblist& collect (list (car lab))))))

(de &macrosubst (mac alist)
  %? Pattern matchers?
  (cons (car mac) (&macrosubst1 (cdr mac) alist)))

(de &macrosubst1 (args alist)
  %? pattern Matchers?
  (foreach arg in args collect (&macrosubst2 arg alist)))

(de &macrosubst2 (arg alist)
  %? Pattern matchers
  (prog (u)
        (setq u (assoc arg alist))
        (return (cond (u (cadr u))
                      ((or (atom arg) (flagp (car arg) 'terminal)) arg)
                      (t (cons (car arg) (&macrosubst1 (cdr arg) alist)))))))

(de &remtags ()
  % Remove auxiliarry tags for old compiler model
  (foreach mac in codelist& do (&remtags1 mac)))

(de &remtags1 (mac)
  % Suppport REMTAGS
  (progn
    (when (equal (car mac) '*jumpon)
      (rplacd (cdddr mac) (list (cddddr mac))))
    (foreach macfield in (cdr mac) do (&remtags2 macfield))))

(de &remtags2 (u)
  % Support REMTAGS
  (when (eqcar u 'wconst)
    (&remtags3 (cadr u))))

(de &remtags3 (u)
  % Support REMTAGS
  (prog (dofn)
        (when (atom u)
          (return nil))
        (when (setq dofn (get (car u) 'dofn))
          (rplaca u dofn))
        (&remtags4 (cdr u))))

(de &remtags4 (u)
  % Support REMTAGS
  (foreach x in u do (&remtags3 x)))

%-------------Entry points used in setting up the system
%--- to declare in COMP-DECLS
(de &onereg (u)
  % Indicate these funs use only one reg
  (foreach x in u do (put x 'destroys '((reg 1)))))

(de &tworeg (u)
  % Indicate these funs use only two reg
  (foreach x in u do (put x 'destroys '((reg 1) (reg 2)))))

(de &threereg (u)
  % Indicate these funs use only three reg
  (foreach x in u do (put x 'destroys '((reg 1) (reg 2) (reg 3)))))

