%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXK:MAIN-START.SL
% Title:          First routine called on startup
% Author:         Brian Beach, Hewlett-Packard CRC
% Created:        16-Feb-84
% Modified:       May 16 1985
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 2-Dec-88 (Winfried Neun)
%  Changed according to new calling sequence etc. (see "ZIB" comments)
% 31-Aug-88 (Julian Padget)
%  Changed fluids to globals, since latter needed in CPSL.
% 28-Oct-87 (Harold Carr)
%  Made panic exit call exit-with-status, rather than redundantely doing
%  what panic exit already did then calling external_exit.
% 13-Apr-87 (Harold Carr & Leigh Stoller)
%  Added external fluid _infbitlength_ to be used in bpsheap.c.
% 26-Sep-86 (Leigh Stoller)
%  Changed exit to unix to be a call to exit library function. The Sun does
%  accept the exit status in (reg 1) like the Bobcat. Called it external_exit
%  so it does not interfere with the cmacro of the same name.
% 22-Sep-86 (Leigh Stoller)
%  Commented out calls 68020_advise which is not defined on the sun.
% 18-Jul-86 (Leigh Stoller)
%  Changed the allocation of argumentblock to be the number of maxregs since
%   it is indexed by base + reg number.
% 15-Feb-85 13:27:53 ()
%  Made it so that PSL allocates the stack it will use.
%  Removed TST.B for stack expansion.
%  Included declaration of STACKSIZE (should go in SYS-CONSTS eventually).
% 9-Jul-84 09:11:55 (RAM)
%  Added fluids (and initialization) unixstdin, unixstdout, unixstderr,
%  unixnull, unixeof, unixtty.
% 3-Jul-84 10:51:27 (RAM)
%  Changed name of initial faslin file to $pnkl/load-psl.b since
%  binaryopenread can now expand shell variable references.
% 29-Jun-84 12:33:38 (RAM)
%  Chagged exit_to_unix to exit-to-unix to match call in kernel-fatal-error.
%  Removed save and restore of A6.
% 31-May-84 08:15:32 (Brian Beach)
%  Changed SYSCLEARIO call to HP_CLEARIO.  SYSCLEARIO is now called by the 
%  loader as part of its startup.
% 16-may-88 (chris burdorf and T. Yamamoto)
%  modified the main entry point so that it could handle our different
%  stack pointer and the sun-4's register windows to get arg c and argv from
%  the operating system.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime    %% Move this hack somewhere else later.
(progn
(OFF LAPOPT)
(deflist
  '((r0  "g0")(r1  "g1")(r2  "g2")(r3  "g3")(r4  "g4")
    (y "y") %
    (i0 "i0")(i1 "i1")(i2 "i2")(i3 "i3")(i4 "i4")(i5 "i5")(i6 "i6")(i7 "i7"))
  '*register-print-name)

(put 'y '*register-code 815)

(de initializesymboltable ()
  (let ((maxsymbol (compiler-constant 'maxsymbols)))
    (when (lessp maxsymbol nextidnumber*)
      (errorprintf "*** MaxSymbols %r is too small; at least %r are needed"
                   maxsymbol nextidnumber*)
      (setq maxsymbol (plus nextidnumber* 100)))
    (flag '(nil) 'nilinitialvalue)
    (put 't 'initialvalue 't)
    (setq nilnumber* (compileconstant nil))
    (dataalignfullword)
    (initializesymfnc)
    (datareservefunctioncellblock
     (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymval)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymprp)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymnam)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymget)   % SYMGET feature
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    ))
(put 'wquotient 'opencode '((sra (reg 1) 31 (reg 4))
                            (wr (reg 4) (reg 0) (reg y))
                            (sdiv (reg 1) (reg 2) (reg 1))))
(put 'wremainder 'opencode '((sra (reg 1) 31 (reg 4))
			    (wr (reg 4) (reg 0) (reg y))
			    (SDIV (reg 1) (reg 2) (reg 3))
			    (smul (reg 3) (reg 2) (reg 4))
			    (*wdifference (reg 1) (reg 4))))

(de dataprintundefinedfunctioncell (idn)
  (prog (oldout)
        (setq oldout (wrs dataout*))
        (dataprintf " .word undefinedfunction %n")
        (wrs oldout)))

(de dataprintdefinedfunctioncell (x idn)
  (datadeclareexternal x)
     % dont disturb undefined function and interpreter
   (dataprintf definedfunctioncellformat* x (id2int idn))))

(setq definedfunctioncellformat!* " .word %w%n")

(setq undefinedfunctioncellinstructions!*
      '((!.word undefinedfunction)
        (sethi !%w (reg o0))))
))

(compiletime    %% Move this hack somewhere else later.
 (defmacro exported-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'exportedsymbol)
         (put id 'symbol id)
         (put id 'vartype 'global)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro internal-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'internalsymbol)
         (put id 'symbol id)
         (put id 'vartype 'global)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro system-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (put id 'vartype 'global)
         ))
   ))

(on fast-integers)

(compiletime (load sys-consts))

(declare-kernel-structure 
 (stack             2 ) %% #.(compiler-constant 'stacksize))
 (argumentblock     #.(compiler-constant 'maxargs))
 (tokenbuffer       #.(compiler-constant 'maxtokensize))
 (bndstk            #.(compiler-constant 'bndstksize))
 (catchstack        #.(times2 (compiler-constant 'catchstacksize) 4))
 (hashtable         #.(quotient (compiler-constant 'hash-table-size) 2))
 (vivarea           2048)
 (onewordbuffer     1)
 (saveargc          1)
 (saveargv          1)
 (datebuffer        1)
%******************************************************
%       SPARC specific. T.Yamamoto, July 12,1988
%
 (sparcjumpbuf          6)
%******************************************************
 )

(exported-global
 '(bndstkptr   bndstkupperbound   gcarraylowerbound   gcarrayupperbound
   heaplast   oldheaplast
   heaplowerbound  heaptrapbound   oldheaptrapbound   heapupperbound
   _infbitlength_   lastbps   bpslowerbound   mainstartinitialize
   nextbps   oldheapupperbound   oldheaplowerbound
   stacklowerbound   stackupperbound   saveda7
   unixstdin   unixstdout   unixstderr   unixnull   unixeof   unixtty
   ))

(internal-global  '(   symnam  symval  symget symfnc  symprp   ))

(system-global  '(symnam symval symprp  symfnc symget
                  %  stack
                  catchstackptr   catchstack
                  bndstklowerbound   bndstkupperbound
                  bndstkptr   bndstk
                  heap
                  nextsymbol    %% Had SYMBOL as its assembler symbol.  why????
                    % Add these so they are always availbale to compiled code.
                  ***MUST-BE-NIL***   ***MUST-BE-NIL-TOO***
                  *fastcar
        %******************************************************
        %       SPARC specific  T.Yamamoto, July 12,1988
        %
                 vivarea    % ZIB
                 sparcjumpbuf
        %******************************************************
   ))

(compiletime (copyd 'saveZIBanyregGlobal 'ZIBanyregGlobal))
(compiletime (de ZIBanyregGlobal (reg source) (list '$GLOBAL source)))

(de init-pointers()

  (setq catchstackptr catchstack)
  (setq bndstklowerbound bndstk)
  (setq bndstkupperbound (loc (wgetv bndstk (wdifference bndstksize 1))))
  (setq bndstkptr bndstk)
  (setq heap heaplowerbound)
  (init-viv))

(de init-fluids-old ()
        % Until problem of labels not appearing is fixed...
  (setq gcarraylowerbound nil)
  (setq gcarrayupperbound nil)
  (setq heaplast nil)
  (setq oldheaplast nil)
  (setq heaptrapbound nil)
  (setq oldheaptrapbound nil)
  (setq heapupperbound nil)
  (setq _infbitlength_ nil)
  (setq lastbps nil)
  (setq bpslowerbound nil)
  (setq mainstartinitialize nil)
  (setq nextbps nil)
  (setq oldheapupperbound nil)
  (setq oldheaplowerbound nil)
  (setq stackupperbound nil)
  (setq unixstdin nil)
  (setq unixstdout nil)
  (setq unixstderr nil)
  (setq unixnull nil)
  (setq unixeof nil)
  (setq unixtty nil)
  (setq ***MUST-BE-NIL*** NIL)
  (setq ***MUST-BE-NIL-TOO*** NIL)
  (setq *fastcar nil))

(compiletime % copying the actual known viv variable to viv area
(progn
 (fluid '(collected-viv*)) 
 
 (dm collect-viv (u) 
   (prog (hhh) 
      (mapobl (function collect-viv-1)) 
      (setq hhh (append `((*entry init-viv expr 0) (*alloc 0)) 
                           (append collected-viv* '((*exit 0)))))
      (return (setq hhh (list 'lap (list 'quote hhh)))))) 
 
 (de collect-viv-1(x) 
    (prog (g) 
      (when (and (setq g (get x 'viv-number)) 
                 (or (get x 'symbol) (get x 'idnumber)))
          (return 
           (setq collected-viv*
            (append `((*movex ($fluid ,x) (reg t5))
                      (*move (reg t5) (indexed (reg g7) ,g))) collected-viv*)))) 
)) 
))

(collect-viv) %do it now

(compiletime (copyd 'ZIBanyregGlobal 'saveZIBanyregGlobal))

(de init-fluids ()
        % Until problem of labels not appearing is fixed...
  (init-fluids-old) % do you know the difference?
  (setq gcarraylowerbound nil)
  (setq gcarrayupperbound nil)
  (setq heaplast nil)
  (setq oldheaplast nil)
  (setq heaptrapbound nil)
  (setq oldheaptrapbound nil)
  (setq heapupperbound nil)
  (setq _infbitlength_ nil)
  (setq lastbps nil)
  (setq bpslowerbound nil)
  (setq mainstartinitialize nil)
  (setq nextbps nil)
  (setq oldheapupperbound nil)
  (setq oldheaplowerbound nil)
  (setq stackupperbound nil)
  (setq unixstdin nil)
  (setq unixstdout nil)
  (setq unixstderr nil)
  (setq unixnull nil)
  (setq unixeof nil)
  (setq unixtty nil)
  (setq ***MUST-BE-NIL*** NIL)
  (setq ***MUST-BE-NIL-TOO*** NIL)
  (setq *fastcar nil)
)

(compiletime (on mainfound))

(lap '((*entry !p!s!l!_!m!a!i!n expr 0)
       (*alloc 1) 
%%ZIB%% (save (reg o6) -256 (reg o6))  % required when we call a foreign func.
%      (*move 10 (reg o1))
%      (sll (reg o1) 23 (reg o1))  % set floating point status register
%      (*move (reg o1) (fluid argc)) % to detect zero-divide, underflow and
%      (*movex (fluid argc) (reg fsr)) % overflow
       (*move 4711 (reg l0))
       (*mkitem (reg l0) forward-tag)

    % We had to change this code here a bit for the SUN-4, because
    % the sun-4 uses register windows to pass arguments and not a stack
    % as on the vax and 68xxx.  

       (*move  (reg i0) (fluid argc))  % system dependent for sun-4. We move
                                       % the argc value (reg i0) into the
                                       %fluid argc. 
       (*move  (reg i1) (fluid argv))  %system dependent for sun-4. move
                                       % the argv value into the fluid argv.

       (*move (reg i0) (reg o0))
       (*move (reg i1) (reg o1))
       (*move (reg i2) (reg o2))  % envp
       (call (foreignentry copy_argv))     % call c function to store argv 
                                           % into static space.
       (nop)
       (*move (reg o0) (fluid argv))
%*****************************************************************************
%       SPARC specific.
%       Set g7 to the vivarea
       (*move (fluid vivarea) (reg r7)) % ZIB
       (add (reg r7) 4092 (reg r7))     % place pointer into middle of array
	(*movex ($fluid SYMFNC) (reg t5))
	(*move (reg t5) (indexed (reg g7) (minus 132)))
       (*move (reg r7) (reg 1))
	(*link preset-viv expr 0)
        (*movex ($fluid SYMFNC) (reg t5))
        (*move (reg t5) (indexed (reg g7) (minus 132)))
       (*move (reg r7) (fluid PSLDAT))  % using r7 now!!
%****************************************************************************

       %  Do OS specific initializations (uses argc and argv)
       (*move (fluid argc) (reg 1))
       (*move (fluid argv) (reg 2))
       (*move infbitlength (fluid _infbitlength_))
       (*link os_startup_hook expr 2)
       
       (*call init-pointers)

       (*move 128 (reg g6))
       (*mkitem (reg g6) id-tag)                    % initialize NIL reg
       (*move 116 (reg 1))                          % initialize T in Vivarea
       (*mkitem (reg 1) id-tag)
       (*move (reg 1) (indexed (reg g7) -36)) %%% ZIB !!
       (*move (reg 1)(fluid !t))
       (*call pre-main)                                 % call PSL

panic-exit                      % need to do UNIX cleanup after
                                % a fatal error, so jump here
       (*move (wconst 0) (reg 1))
       (*linke 1 exit-with-status expr 1)
       (*exit 1)

       (*entry exit-with-status expr 1)
         (*move (reg 1) (fluid onewordbuffer))
         (*link os_cleanup_hook expr 0)
         (*move (fluid onewordbuffer) (reg 1))
         (*link external_exit expr 1)
         (*exit 1)))

(de pre-main ()
  (unixcleario)
  (initialize-symbol-table)
  (initcode)
  (console-print-string "About to load LOAD-PSL")
  (console-newline)
  (faslin "load-psl.b")
  (loader-main))

(de preset-viv (p)
  (for (from i 0 -512 -1) 
     (do (wputv p i (mkitem unbound-tag 'preset-viv)))))

(compiletime (when (getenv "linux")
(lap '((*entry fpsetmask expr 1)
	(save (reg st) -256 (reg st))
	(sethi (hi 16#1800000) (reg o4))
	(sll (reg 1) 16#17 (reg 2))
	(st  (reg fsr) (displacement (reg st) 16#44))
	(ld (displacement (reg st) 16#44) (reg 1))
	(and (reg 2) (reg o4) (reg 1))
	(andn (reg 1) (reg o4) (reg 3))
	(or (reg 2) (reg 3) (reg 2))
	(st (reg 2) (displacement (reg st) 16#44))
	(ld (displacement (reg st) 16#44) (reg fsr))
        (and (reg 1)  (reg o4) (reg 1))
	(srl  (reg 1) 16#17 (reg 1))
	(ret)
	(restore)))

(lap '((*entry fpsetsticky expr 1)
       (save    (reg st) -96 (reg st))
       (mov     16#3e0 (reg 5))
       (sll     (reg 1) 5 (reg 2))
       (st      (reg fsr) (displacement (reg st) 16#44))
       (ld      (displacement (reg st) 16#44) (reg 1))
       (and     (reg 2) (reg 5) (reg 2))
       (andn    (reg 1) (reg 5) (reg 3))
       (or      (reg 2) (reg 3) (reg 2))
       (st      (reg 2) (displacement (reg st) 16#44))
       (ld      (displacement (reg st) 16#44) (reg fsr))
       (and     (reg 1) (reg 5) (reg 1))
       (srl     (reg 1) 5 (reg 1))
       (ret)
       (Restore)))

% Must live in the kernel in order to restore g6 and g7

(lap '((*entry signalhandler expr 0)
       (*alloc 0)
       (*move (reg o0) (reg 1))
       (*move (reg o1) (reg 2))
       (*move (reg o2) (reg 3))
       (*move (reg o3) (reg 4))
       (*move (reg o4) (reg 5))
       (*move (fluid vivarea) (reg r7))
       (add (reg r7) 4092 (reg r7))     % place pointer into middle of array
       (*move 128 (reg g6))
       (*mkitem (reg g6) id-tag)                    % initialize NIL reg
       (*linke 0 signalhaendler expr 0)))

))
(off fast-integers)
