%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:COPYING-GC.SL 
% Description:  Copying 2-space garbage collector for PSL 
% Author:       Eric Benson 
% Created:      30 November 1981 
% Modified:     29-Aug-84 10:09:08 (Brian Beach) 
% Package:      
%
% (c) Copyright 1987, University of Utah, all rights reserved.
% (c) Copyright 1989, Konrad Zuse Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%  28-Sep-93 (Winfried Neun)
%  Added support for 256 Fixed ids
%  21-Jul-89 (Winfried Neun)
%  Modified code to support dynamic expansion of heap.
%  10-Mar-89 (Winfried Neun)
%  Using more heap space now than 50%.
%  1-Jan-89 (Winfried Neun)
%  Changed handling of stack contents for 'new' Sparc PSL
%  Replaced Copyitem and Copyitem1 by a new Copyitem using Case clause
%  and overlapping load.
%  1-Sep-88 (Julian Padget)
%  Changed system fluid variables to global.
% 16-Jun-87 (Leigh Stoller)
%  Added Mark Swanson's hack to the GC that ignores bogus stack items.
% 10-Dec-1986 (Tim Mueller)
%  Fixed bug in reclaim2() where oldHeapLowerBound
%   might be equal to heapUpperBound (changed > to >=).
% 9 Oct 1986 - Fix by Russ Fish and Tim Mueller to eliminate device dependent
% code in reclaim2 where null objects are detected.  Also can now repeat
% reclaim2's.
% 6-Feb-86 08:10:09 (Mark Swanson)
%  put in some NIL checks (strictly) for speed;  took out a gratuitous
%  putmem in the non-pointertag case of copyitem
% 6-Feb-86 07:15:59 (Mark Swanson)
%  changed calls on copyfrombase to copyitem, since that's all former
%  did was to call the latter anyway
% 5-Feb-86 21:54:00 (Mark Swanson)
%  Added RECLAIM2 for static heap from 3.2 implementation.
% 24-Jul-84 08:29:01 (Brian Beach)
%  Changed names of tags for micro-kernel.
%  Removed uses of CONST.
% 3 Jul 1984 (Vicki O'Day)
%  Changed references to the constant forward to forward-tag, to match
%  the declaration in tags
% 26 Jan 1984 0958-PST (Brian Beach)
%  Corrected file name in header.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Old Revision history:
%
% 14 Dec 83 (Cris Perdue)
%  Added handling of BYTES and HALFWORDS to this version.
% Edit by Mark Swanson, 3 April 1983 0949-MST
% Made COPYITEM1 tail-recursive so that long lists can be copied without
% stack overflow -- both COPYITEM and COPYITEM1 are now called exclusively
% for their side effects--return values should be ignored.
% Edit by Cris Perdue, 25 Mar 1983 1711-PST
% Occurrence of heap-warn-level in initialization no longer flagged
% with "LispVar".  Didn't work.
%  <PSL.KERNEL>COPYING-GC.RED.2, 23-Mar-83 11:35:37, Edit by KESSLER
%  Add HeadTrapBound Guys, so we can update the heap trap bound upon switch
% Edit by Cris Perdue, 15 Mar 1983 0937-PST
% Added missing comma as noted by Kessler.
% Edit by Cris Perdue, 16 Feb 1983 1409-PST
% Removed external declaration of HeapPreviousLast (the only occurrence)
% Now using "known-free-space" function and heap-warn-level
% Sets HeapTrapped to NIL now.
% Added check of Heap!-Warn!-Level after %Reclaim.
%  <PSL.KERNEL>COPYING-GC.RED.6,  4-Oct-82 17:56:49, Edit by BENSON
%  Added GCTime!*

(on fast-integers)

(compiletime (load hash-decls if-system))

(compiletime (flag '(moreheap) 'internalfunction))

(fluid '(PSLDAT SYMGET *bruch* *gc gcknt* gctime* heap heapenlarge*
	 heap-warn-level !*dump))

(setq heapenlarge* 250000)

(loadtime
  (progn (setq gcknt* 0)
         (setq gctime* 0)
         (setq StaticHeapBase 0)
         (setq StaticHeapUpperBound 0)
         (setq *gc t)
	 (setq heap-warn-level 30000)))

(compiletime
 (progn
  (ds !%chipmunk-kludge (x) x)   %%% TEMPORARY

%  (ds pointertagp (x)
%    (and (> x posint-tag) (< x code-tag)))

  (ds pointertagp (x)
    (wgeq (wor (- x (wplus2 posint-tag 1)) (- (wplus2 code-tag -1) x)) 0))
                                   % only 1 jump version

  (ds itemtagp (x)
    (or (eq x id-tag)
        (eq x negint-tag)
        (eq x code-tag)
        (<= x evector-tag)))

%  (ds withinoldheappointer (x)
%    (and (wgeq x (!%chipmunk-kludge oldheaplowerbound))
%         (wleq x (!%chipmunk-kludge oldheaplast))))    

  (ds withinoldheappointer (x)
     (wgeq (wor (- x oldheaplowerbound)
                (- oldheaplast x))
           0))    

  (ds withinlocaloldheappointer (x)
     (wgeq (wor (- x localoldheaplowerbound)
                (- localoldheaplast x))
           0))

  (ds mark (x)
    (mkitem forward-tag x))

  (ds marked (x)
    (weq (tag x) forward-tag))

  (ds markid (x)
    (setf (field (symnam x) tagstartingbit tagbitlength)
      forward-tag))

  (ds markedid (x)
    (= (tag (symnam x)) forward-tag))

  (ds clearidmark (x)
    (setf (field (symnam x) tagstartingbit tagbitlength) string-tag)
    )

  (ds exchange (a b)
    (let ((*temp* a))
      (setf a b)
      (setf b *temp*)
      ))
  
  (flag '(copyfromallbases copyfromstaticheap copyffromonestaticheap
          copyfromrange    current-stack-pointer
          copyfrombase     copyfromstack
          copyitem         copyfromstack1


%  copyitem1
          markandcopyfromid copyfromstack11
          makeidfreelist    cfserror
          gcstats           copyfromonestackitem)
        'internalfunction)
  ))

(global '(heaptrapped
         oldheaptrapbound
         heaptrapbound
         oldheapupperbound
         oldheaplowerbound
         oldheaplast
         heapupperbound
         heaplowerbound
         heaplast
         bndstkptr
         bndstklowerbound
         stacklowerbound
%        st
         nextsymbol
         StaticHeapBase
         StaticHeapUpperBound
         ))

(fluid '(oldsize 
         oldtime
%%       stacklast
         *grosso-gc*
         ))

(setq heaptrapbound (wdifference heaptrapbound 120))
(setq oldheaptrapbound (wdifference oldheaptrapbound 120))

(commentoutcode (de Reclaim2 nil 
  (prog (B oldhsize newsize staticsize) 
    (cond ((geq OldHeapLowerBound HeapUpperBound) 
        (!%Reclaim)))                   % make sure we're in upper heap
    (!%Reclaim)                         % to get into lower heap
    (setq oldhsize (difference HeapUpperBound HeapLowerBound)) 
    % Only set this once (may be called more than once)
    (when (eq StaticHeapBase 0)
          (setq StaticHeapBase HeapLowerBound))
    (cond ((equal (land HeapLast 7) 4) 
        (setq HeapLast (plus HeapLast 4)))) 
    (setq HeapLowerBound HeapLast) 
    (setq StaticHeapUpperBound HeapLast) 
    (setq staticsize (difference StaticHeapUpperBound StaticHeapBase)) 
    (setq HeapUpperBound (plus HeapLowerBound 
        (quotient (difference OldHeapUpperBound HeapLowerBound) 2))) 
    (setq OldHeapLowerBound (plus HeapUpperBound 4)) 
    (setq newsize (difference HeapUpperBound HeapLowerBound)) 
    (setq HeapTrapBound (wplus2 HeapUpperBound -120)) 
    (setq OldHeapTrapBound (wplus2 -120 OldHeapUpperBound)) 
    (printf "Old Heap Size: %d, New Heap Size: %d, Static Heap Size: %d" 
        oldhsize newsize staticsize)))
)
  
(de reclaim ()
    (prog (b)
  (!%reclaim)
)) 

(de !%reclaim ()
  (prog (b oldtimehi oldtimelo enlargement-tried)
  (when *gc (errorprintf "*** Garbage collection starting"))
  (setq *bruch* 17) % gc should not be interrupted
 another-one
  (beforegcsystemhook)
  (setq *grosso-gc* T)
  (setq oldtimehi (timc)) % a strange problem after 37 hours
  (setq oldtimelo (wand oldtimehi 16#ffff))
  (setq oldtimehi (wshift oldtimehi -16))
  (setq oldsize (wdifference heaplast heaplowerbound))
  (setf gcknt* (+ gcknt* 1))

  (setq oldheaplast heaplast)
  (setq heaplast oldheaplowerbound)

  (exchange heaplowerbound  oldheaplowerbound)    % Swap heaps
  (exchange heapupperbound  oldheapupperbound)
  (exchange heaptrapbound   oldheaptrapbound)

  (copyfromallbases)
  (makeidfreelist)
  (aftergcsystemhook)

  (setq oldtime (wor (wshift oldtimehi 16) oldtimelo))
  (setq oldtime (wdifference (timc) oldtime))
  (setf gctime* (wplus2 gctime* oldtime))
  (when (not (posintp gctime*)) (setq gctime* (inf -1)))
                 % another 37 hours problem
  (when (wlessp *bruch* 0) (lisprpv))
  (when *gc
    (gcstats))
  (setq *bruch* 0)
  (setq heaptrapped nil)

  (when enlargement-tried
    (if (wleq (moreheap heapenlarge*) 0)
        (progn (continuableerror 99 "Heap space low" nil)
               (return NIL))
        (when *gc
            (Errorprintf "*** Heap space has been enlarged by %w items"
                                heapenlarge* ) )
        (when *gc (gcstats))
        (return NIL)
    ))

  (when (or (wlessp (intinf (known-free-space)) (intinf heap-warn-level))
            (force-heap-enlargement))
    % try to enlarge heap space via call to alterheapsize
    % make sure that stuff is in lower part
    (when (wgreaterp heapupperbound oldheapupperbound)
          (setq enlargement-tried t)
          (go another-one))
    (if (wleq (moreheap heapenlarge*) 0)
        (continuableerror 99 "Heap space low" nil)
        (when *gc
            (Errorprintf "*** Heap space has been enlarged by %w items"
                                heapenlarge* ) )
        (when *gc (gcstats))
    ))

  nil
  ))

(de moreheap (incr)
    (setq incr (times2 incr addressingunitsperitem))
    (when (wgreaterp (alterheapsize incr) 0)
      (setq heapupperbound (wplus2 heapupperbound  incr))
      (setq heaptrapbound  (wplus2 heapupperbound -120 ))
      (setq oldheaplowerbound (wplus2 oldheaplowerbound  incr))
      (setq oldheapupperbound (wplus2 oldheapupperbound (wtimes2 incr 2)))
      (setq oldheaplast oldheaplowerbound)
      (setq oldheaptrapbound (wplus2 oldheapupperbound -120))
))

(de force-heap-enlargement() NIL)

(de set-heap-size (items)
   (prog (actual heap-warn-level heapenlarge*)
       (setq actual (difference heapupperbound heaplowerbound))
       (setq actual (quotient actual 4))
       (when (null items) (return actual))
       (when (greaterp actual items)
          (print "Cannot shrink heap") (return nil))
       (setq heapenlarge* (difference items actual))
       (setq heap-warn-level actual)
       (when (pairp (errorset '(!%reclaim) NIL NIL)) (return items))
       (return nil)))

(de set_heap_size (items)
  (set-heap-size items))

(if_system sel
(lap '((*entry current-stack-pointer expr 0)
       (*move (reg 0) (reg t1))
       (*move (fluid nil) (reg nil))% Save return address and set stack pointer
                                    % as it will be after the return.
       (*move (reg st) (reg 1))
       (*jump (indexed(reg t1) 0))
       ))

(lap '((*entry current-stack-pointer expr 0)
       (*move (reg st) (reg 1))
       (jmpl (indexed (reg o7) 8) (reg g0))
       (nop)
       ))
)

(de markandcopyfromid (x)
  % SymNam has to be copied before marking, since the mark destroys the tag
  % No problem since it's only a string, can't reference itself.
%(when (wgreaterp x 32)(console-print-string (symnam x)) (console-newline))

  (progn (copyitem (loc (symnam x)) (symnam x)) 
         (markid x)
         (copyitem (loc (symprp x)) (symprp x))
         (copyitem (loc (symval x)) (symval x))
         (copyitem (loc (symget x)) (symget x))
))

(de copyfromallbases ()

  % Mark fixed ID's;  NIL first.
  (markandcopyfromid 128)
  (for (from i 0 127 1)
       (do (unless (markedid i)
             (markandcopyfromid i))))

  (for (from i 129 255 1)
       (do (unless (markedid i)
             (markandcopyfromid i))))

  % Mark other IDs.
 (if *grosso-gc*
  (for (from i 0 hash-table-size 1)
       (do
        (let ((id-number  (hash-table-entry i)))
          (when (and (> id-number 0)     % Slot is occupied
		     (< id-number 16#FFFF)
                     (not (markedid id-number)))
            (markandcopyfromid id-number)
            ))))
% marking all ids
  (for (from i 0 maxsymbols 1)
      (do (when (stringp (symnam i))
            (markandcopyfromid i))))
)

  % Mark from binding stack.
  (for (for b (adjustbndstkptr bndstklowerbound 1) (adjustbndstkptr b 1))
       (while (wleq b bndstkptr))
       (do (copyitem b (getmem b)))
       )

  (for (from b (wdifference psldat 512) psldat 4)
       (do (copyitem b (getmem b)))) % copy from VIV area
        
  (CopyFromStack)
  (CopyFromStaticHeap))

% Stack start is indicated by a forward/0815 item and stackend by a
% forward/4711 item (rather German numbers indeed)
% We must set the 0815 item before windows are rolled out.

(lap '((*entry xxcopyfromstack expr 0)
       (*alloc 0)
       (*move 0815 (reg l0))
       (*mkitem (reg l0) forward-tag)
       (*move 1 (reg 1))
       (*link copyfromstack1 expr 1)
       (*exit 0)))

(compiletime (put 'get-stack-hack 'opencode '((*move (reg st) (reg 1)))))

(de copyfromstack() (copyfromstack1 1 (get-stack-hack)))

(de copyfromstack1 (i stpointer)
   (if (wgreaterp i 31) (copyfromstack11 stpointer)
                 % writing register windows out to memory
        (copyfromstack1 (wplus2 i 1) stpointer))
        17) % avoid tail recursion

(de copyfromstack11 (stpointer)
  (prog (x validitems nextframe fwd starting ending )
    (setq starting (get-stack-hack))
    (setq ending 16#ffff) % used to be f7ff is a hack anyway
    (setq ending (wor (wshift ending 16) 16#ffff))
%    (while (and (wleq starting ending) % DO NOT construct the item
%                (or (not (eq (tag (getmem starting)) forward-tag))
%                    (not (eq (inf (getmem starting)) 815))))
%          (setq starting (wplus2 starting
%               (wminus (wtimes2 addressingunitsperitem stackdirection)))))
                                            % we found top of stack
    (setq starting stpointer) % Versuch
    (setq fwd 96) % its a magic number for sparc
  lupe                                      % loop on stack frames
%   (setq nextframe (wplus2 starting FWD))
    (setq nextframe (wgetv starting 14)) % backward (frame) pointer
    (setq starting nextframe)
%   (when (wgreaterp starting ending) (return (print "ueberszielhinaus")))
 
    (setq validitems (getmem starting))

    (when (not (eq (tag validitems ) forward-tag)) 
          (when *dump  (print "*** Irregular Stack***"))
          (go lupe))
    (when (eq (inf validitems) 4711) (return nil)) % all done

    (setq validitems (wshift (inf validitems) -10))
%   (unixputn (inf validitems)) (console-print-string "-----------")
%   (console-newline)
 
% we walk through the stack and PRE-DELETE all stack positions which are
% not included in the LISP frame (There is a difference between Sparc frame
% and LISP frame!

    (if (wleq validitems 12)
       (progn
        (for (from i 16 23 )
                (do (wputv starting  i (mkitem id-tag (wplus2 starting 13)))))
        (for (from i (wplus2 validitems 1) 12) (do
                   (wputv starting  i (mkitem id-tag (wplus2 starting 15)) ))) 
        (for (from i 1 validitems)     % for all allocated frame positions
            (do (copyfromonestackitem (wplus2 starting (wtimes2 i 4)))))
       ) % a large frame
        (for (from i 16 24 ) 
          (do (wputv starting  i (mkitem id-tag (wplus2 starting i)))))
        (if (eq (wand validitems 1) 0) % even case
            (wputv starting (wplus2 validitems 13)
                      (mkitem id-tag (wplus2 starting 17)))
            % uneven case
            (wputv starting (wplus2 validitems 13)
                      (mkitem id-tag (wplus2 starting 19)))
            (wputv starting (wplus2 validitems 14)
                      (mkitem id-tag (wplus2 starting 21))))
        (for (from i 1 12)
            (do (copyfromonestackitem (wplus2 starting (wtimes2 i 4)))))
        (for (from i 13 validitems)     % for all allocated frame positions
            (do (copyfromonestackitem 
                     (wplus2 48 (wplus2 starting (wtimes2 i 4)))))))

%   (setq fwd                      % forward in stack
%       (cond ((wlessp validitems 13) 96)   % formula from *alloc
%             ((eq 0 (wand validitems 1)) (wtimes2 4 (wplus2 14 validitems)))
%             (t (wtimes2 4 (wplus2 15 validitems)))))
    (go lupe)))

(de copyfromonestackitem (i)
   (prog (tg x info tg2 tg3 NullSize Size stacklast ending)
      (setq NullSize (Inf -1 )) % The size of a null vector.
      (setq x (getmem i))
      (setq tg (tag x))
%      (unixputn x)
      (setq info (inf x))
      (if (and (weq tg id-tag) (wlessp info maxsymbols) (not (markedid info)))
                  (markandcopyfromid info)
          (if (and (withinoldheappointer info) (pointertagp tg)
              (eq (wand info 3) 0)) % no BUS error
            (progn
                (setq tg2 (tag (getmem info)))
                (if (weq tg2 forward-tag)
                  (setf (getmem i) (mkitem tg (inf (getmem info))))
                  (setq Size (Inf (getmem info)))
                  (case tg
                        ((fixnum-tag floatnum-tag words-tag)
                         (if (and (eq tg2 hwords-tag)
                                  (or (eq Size NullSize)
                                      (wleq (+ info (* Size
                                                   addressingunitsperitem))
                                            oldheaplast)))
                           (copyitem i (getmem i))
                           (cfserror "words/fixnum/float" x i)
                           ))
                        ((string-tag bytes-tag)
                         (if (and (eq tg2 hbytes-tag)
                                  (or (eq Size NullSize)
                                      (wleq (+ info Size) oldheaplast)))
                           (copyitem i (getmem i))
                           (cfserror "string" x i)
                           ))
                        ((halfwords-tag)
                         (if (and (eq tg2 hhalfwords-tag)
                                  (or (eq Size NullSize)
                                      (wleq (+ info (* Size 2))
                                            oldheaplast)))
                           (copyitem i (getmem i))
                           (cfserror "halfword" x i)
                           ))
                        ((vector-tag evector-tag)
                         (if (and (eq tg2 hvect-tag)
                                  (or (eq Size NullSize)
                                      (wleq (+ info (* Size
                                                       addressingunitsperitem))
                                            oldheaplast)))
                           (copyitem i (getmem i))
                           (cfserror "(e)vector" x i)
                           ))
                        ((bignum-tag)
                         (if (and (or (eq tg2 hwords-tag) (eq tg2 hvect-tag))
                                  (or (eq Size NullSize)
                                      (wleq (+ info (* Size
                                                       addressingunitsperitem))
                                            oldheaplast)))
                           (copyitem i (getmem i))
                           (cfserror "bignum" x i)
                           ))

                        ((pair-tag)
                         (if (itemtagp tg2)
                           (progn
                               (setq tg3 (tag
                                          (getmem
                                           (wplus2 info
                                                   addressingunitsperitem))))
                               (when (itemtagp tg3)
                                     (copyitem i (getmem i))))
                           (cfserror "pair" x i)))
                        (nil (cfserror "nil" x i)))
                  ))))
    %   (console-print-string " | ") (unixputn (getmem i))
    %   (console-print-string " | ")(unixputn i)
    % (console-newline)
))

(de cfserror (where stackitem pos)
   (prog (info hinfo htag outputbase*)
         (setq outputbase* 16)
         (setq info (inf stackitem))
         (setq hinfo (getmem info))
         (setq htag (tag hinfo))
	(when !*dump
         (when *gc (centerdump (inf stackitem) 10))
         (when *gc (centerdump pos 20))
         (errorprintf
          "CopyFromStack error in %w, stackitem %w heaptag %w heapinf %w%n" 
                 where (inf stackitem) htag (inf hinfo)))))

(de CopyFromStaticHeap nil 
(prog (B Size NullSize) 
    (setq NullSize (Inf -1 ))   % The size of a null vector.
    (setq B StaticHeapBase) 
    (while (wlessp B StaticHeapUpperBound) 
    (progn (setq Size (Inf (getmem B))) 
        (case (Tag (getmem B))
        ((hwords-tag) 
        (progn (cond ((equal Size NullSize) 
              (setq Size -1))) 
            (setq B (wplus2 B (<< Size 2) 8)))) 
        ((hbytes-tag) 
        (progn (cond ((equal Size NullSize) 
              (setq Size -1))) 
            (setq B (wplus2 B (<< (>> (wplus2 Size 5) 2) 2) 4)))) 
        ((hhalfwords-tag) 
        (progn (cond ((equal Size NullSize) 
              (setq Size -1))) 
            (setq B (wplus2 B (<< (>> (wplus2 Size 2) 1) 2) 4)))) 
        ((hvect-tag) 
        (progn (cond ((equal Size NullSize) 
              (setq Size -1))) 
            (setq B (wplus2 B 4)) 
            (for (from N 0 Size 1) (do 
            (progn (CopyItem B (getmem b)) 
                (setq B (wplus2 B 4)))))))
        (nil    
        (progn (CopyItem B (getmem b)) 
            (setq B (wplus2 B 4)) 
            (cond ((getmem B) (CopyItem B (getmem b))))
            (setq B (wplus2 B 4)))))))))

(de copyfromrange (lo hi)
  (prog (x i)
        (setq x lo)
        (setq i 0)
        (while (wleq x hi)
          (copyfrombase x)
          (setq i (wplus2 i 1))
          (setq x (loc (wgetv lo i))))))

(de copyfrombase (p)
  % P is an "address"
  (copyitem p (getmem p)))

(commentoutcode de copyitem (p x)
  % not a clean optimization:
  %    typ info and hdr are NOT passed as args, but they should not
  %    preset to NIL, since they are not used as NIL

   (when (null x) (return p))
   (prog (typ info hdr strips news ptr len)
%      (setq x (getmem p))
%      (when (null x) (return nil))
       (setq typ (tag x))
       (case typ
           ((posint-tag negint-tag code-tag btr-tag unbound-tag
             bvector-tag context-tag funarg-tag sgd-tag bstruct-tag)
                         (return nil))
           ((id-tag) 
               (progn (setq info (idinf x))
                  (return (unless (markedid info) (markandcopyfromid info)))))
           ((pair-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
%                    (rplaca x (mark (setq news (gtheap (pairpack)))))
                     (rplaca x (mark (setq news heaplast)))
                     (setq heaplast (wplus2 heaplast 8)) % this is faster
                     (setf (getmem p) (mkpair news))
                     (setf (wgetv news 0) hdr)
                     (setf (wgetv news 1) (cdr x))
                     (copyitem (loc (wgetv news 0)) (wgetv news 0))
                     (return (copyitem (loc (wgetv news 1))(wgetv news 1)))))
          ((bytes-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
                     (setf (getmem info) (mark (setq news (copybytes x))))
                     (return (setf (getmem p) news))))
          ((string-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
                     (setf (getmem info) (mark (setq news (copystring x))))
                     (return (setf (getmem p) news))))               
          ((floatnum-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
                     (setf (getmem info) (mark (setq news (gtfltn))))
                     (setf (wgetv news 1) (wgetv info 1))
                     (setf (wgetv news 2) (wgetv info 2))
                     (return (setf (getmem p) (mkitem floatnum-tag news)))))
          ((vector-tag evector-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
                     (setq strips info)
                     (setq len (veclen strips))
                     (setf (getmem strips) (mark (setq ptr (gtvect len))))
                     (for (from i 0 len 1)
                      (do
                        (progn (setf (vecitm ptr i) (vecitm strips i))
                              (copyitem (loc (vecitm ptr i))(vecitm ptr i)))))
                     (return (setf (getmem p)  (mkitem typ ptr)))))
          ((words-tag fixnum-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
                     (setf (getmem info) (mark (setq news (copywrds x))))
                     (return (setf (getmem p) (mkitem typ news)))))
          ((bignum-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
                     (setf (getmem info) (mark (setq news (copywrds x))))
                     (return (setf (getmem p) (mkitem typ news)))))
          ((halfwords-tag)
             (progn  (setq info (inf x))
                     (unless (withinoldheappointer info) (return x))
                     (setq hdr (getmem info))
                     (when (marked hdr)
                       (return (setf (getmem p) (mkitem typ (inf hdr)))))
                     (setf (getmem info) (mark (setq news (copyhalfwords x))))
                     (return (setf (getmem p) (mkitem halfwords-tag news)))))
          ((forward-tag hvect-tag hhalfwords-tag hbytes-tag 21 19 18 17 16 15)
            (printf "*** In %w: Unexpected tag %w found at %w during gc %n"
                   p typ (inf x))))))

(de makeidfreelist ()
  (prog (previous)
        (for (from i 0 255 1) (do (clearidmark i)))
        (setq previous 256)
        (while (and (markedid previous) (< previous maxsymbols))
          (clearidmark previous)
          (setq previous (+ previous 1)))
        (if (< previous maxsymbols)
	  (setq nextsymbol previous)
          (setq nextsymbol 0))
        (for % free list starts here
              (from i (+ previous 1) maxsymbols 1)
              (do (if (markedid i)
                    (clearidmark i)
                    (progn (setf (symnam previous) i)
                           (setq previous i)))))
        (setf (symnam previous) 0)))
 
% end of free list

(de gcstats ()
  (if (wgeq (known-free-space) 100)
      (if (wlessp gctime* 10000)
        (Errorprintf "*** GC %w: %w (~ %w ms cpu time, gc : %w %%)"
                       gcknt* (date-and-time) (sys2int(timc))
                              (wquotient (wtimes2 100 gctime*) (timc)))
        (Errorprintf "*** GC %w: %w (~ %w ms cpu time, gc : %w %%)"
                       gcknt* (date-and-time) (sys2int(timc))
                              (wquotient gctime* (wquotient (timc) 100))))
       (Errorprintf "*** GC %w: Heap space exhausted" gcknt*))
  (Errorprintf "*** time %d ms, %d occupied, %d recovered, %d free"
     oldtime
     (wquotient (wdifference heaplast heaplowerbound) addressingunitsperitem)
     (if (wlessp (wdifference oldsize (wdifference heaplast heaplowerbound))
	   0)
         0
	 (wquotient (wdifference oldsize (wdifference heaplast
						      heaplowerbound))
                    addressingunitsperitem))
     (known-free-space)))

% for test purposes

(de myaftergcsystemhook()
  (for (from i heaplowerbound (wplus2 -1 heaplast) 4)
     (do (when (eq (tag (getmem i)) forward-tag)
            (console-print-string " Leiche im Landwehrkanal ")
            (unixputn i) (centerdump i 200) 
                         (centerdump (inf (getmem i)) 200)
            )))
  (checkfromid 128)
  (for (from i 0 127 1)
       (do (unless (markedid i)
             (checkfromid i))))

  % Mark other IDs.
  (for (from i 0 hash-table-size 1)
       (do
        (let ((id-number  (hash-table-entry i)))
          (when (> id-number 0)     % Slot is occupied
            (checkfromid id-number)
            ))))
 
  % Mark from binding stack.
  (for (for b (adjustbndstkptr bndstklowerbound 1) (adjustbndstkptr b 1))
       (while (wleq b bndstkptr))
       (do (checkitem b (getmem b)))
       )
 
  (for (from b (wdifference psldat 512) psldat 4)
       (do (checkitem b (getmem b)))) % copy from VIV area
        
  (checkFromStack 1)
)
(fluid '(*show-new-ids))
(de checkfromid (x)

  % SymNam has to be copied before marking, since the mark destroys the tag
  % No problem since it's only a string, can't reference itself.
(when (and *show-new-ids (wgreaterp x 32))
      (console-print-string (symnam x)) (console-newline))

  (progn (checkitem (loc (symnam x)) (symnam x))
         (checkitem (loc (symprp x)) (symprp x))
         (checkitem (loc (symval x)) (symval x))
))

(de checkitem (x) nil)

(de xxcheckitem (i)
   (prog (x)
     (setq x (getmem i))
     (when (eq (tag x) forward-tag)
             (unixputn x) (console-newline)
             (unixputn i) (console-newline)
            (printf " __________> %w %w " x i)
            (unixputn x) (console-newline))
     (when (pairp x) 
            (when
                (and (wlessp (inf x) oldheapupperbound) 
                     (wgreaterp (inf x) oldheaplowerbound)) 
             (unixputn x) (console-newline)
             (unixputn i) (console-newline)
            (printf " __________> %w %w " x i) 
            (unixputn x) (console-newline))
            (checkitem (inf x))
            (checkitem (wplus2 4 (inf x))))
     (when (and (pointertagp (tag x))
                (wlessp (inf x) oldheapupperbound)
                (wgreaterp (inf x) oldheaplowerbound))
            (unixputn x) (console-newline)
             (unixputn i) (console-newline)
            (printf " __________> %w %w " x i)
            (unixputn x) (console-newline))))
 
(de checkitemforward (i)
   (prog (x)    % no test for forward tag from stack
     (setq x (getmem i))
     (when (pairp x) 
            (when
                (and (wlessp (inf x) oldheapupperbound)
                     (wgreaterp (inf x) oldheaplowerbound))
             (unixputn x) (console-newline)
             (unixputn i) (console-newline)
            (printf " __________> %w %w " x i)
            (unixputn x) (console-newline))
            (checkitem (inf x))
            (checkitem (wplus2 4 (inf x))))
     (when (and (pointertagp (tag x))
                (wlessp (inf x) oldheapupperbound)
                (wgreaterp (inf x) oldheaplowerbound))
            (unixputn x) (console-newline)
             (unixputn i) (console-newline)
            (printf " __________> %w %w " x i)
            (unixputn x) (console-newline))))

(lap '((*entry checkfromstack expr 0)
       (*alloc 0)
       (*move 0815 (reg l0))
       (*mkitem (reg l0) forward-tag)
       (*move 1 (reg 1))
       (*link checkfromstack1 expr 1)
       (*exit 0)))

(de checkfromstack1 (i)
   (if (wgreaterp i 31) (checkfromstack11)
                 % writing register windows out to memory
        (checkfromstack1 (wplus2 i 1)))
        17) % avoid tail recursion

(de checkfromstack11 ()
   (prog (x validitems nextframe fwd starting ending)
    (setq starting (get-stack-hack))
    (setq ending 16#f7ff)
    (setq ending (wor (wshift ending 16) 16#ffff))
    (while (and (wleq starting ending) % DO NOT construct the item
                (or (not (eq (tag (getmem starting)) forward-tag))
                    (not (eq (inf (getmem starting)) 815))))
          (setq starting (wplus2 starting
               (wminus (wtimes2 addressingunitsperitem stackdirection)))))
    (setq starting (wplus2 starting 96))
    (setq fwd t)
    (while fwd
      (checkitemforward starting)
      (setq starting (wplus2 starting 4))
      (when (eq (tag (getmem starting)) forward-tag)
        (when (eq (inf (getmem starting)) 4711)
%             (print "stopping")
              (setq fwd nil))))))

(de centerdump (loc howmuch) 
  (console-newline)
  (console-print-string "**** Center dump at 16#")
  (unixputn loc)
  (console-newline)
  (console-newline)
  (for (from  i (minus howmuch) howmuch 8) 
    (do (unixputn (wplus2 loc (wtimes2 4 i)))
        (console-print-string ": ")
        (for (from j 0 7)  
          (do (unixputn (getmem (wplus2 loc (wtimes2 4 (wplus2 i j))))) 
              (console-print-string " "))) 
        (console-newline)))) 
%
% The BIGGC copying-gc; allowing up to 31/32 of the heap to be used
% vs 1/2 in the normal version
% 

(fluid '(*the-heaplowerbounds* *the-heapupperbounds* *number-of-heaps*
         *max-number-of-heaps*
         *the-heaplasts* *first-heap* *actual-heap* *next-heap*))

(setq *the-heaplowerbounds* (gtwarray 32))
  
(setq *the-heapupperbounds* (gtwarray 32))

(setq *the-heaplasts* (gtwarray 32))

(de installBIGGC (x) 
    (prog (size)
      (!%reclaim)
      (when (wlessp oldheaplowerbound heapupperbound) (!%reclaim))
            % make sure we are in lower heap
      (if (or (lessp x 3) (wgreaterp x 32))
           (return (stderror "*** Parameter must be >2 and <33"))
           (setq *max-number-of-heaps* x)
           (setq *number-of-heaps* 2))
      (setq size (wtimes2 2 (wdifference heapupperbound heaplowerbound)))
      (setq size (wand (wquotient size x) (inf -4))) % avoid bus error
      (when (wgreaterp (wdifference heaplast heaplowerbound) size)
            (return (stderror " Cannot splitt beyond actual utilization")))
      (wputv *the-heaplowerbounds* 1 heaplowerbound)
      (wputv *the-heapupperbounds* 1 (wplus2 size heaplowerbound))
      (for (from i 2  x)
         (do  (wputv *the-heaplowerbounds* i
                        (wgetv *the-heapupperbounds* (isub1 i)))
              (wputv *the-heaplasts* i (wgetv  *the-heaplowerbounds* i))
              (wputv *the-heapupperbounds* i
                        (wplus2 size (wgetv *the-heaplowerbounds* i)))))
      (setq *first-heap* 1)
      (setq *next-heap* 2)
      (setq *actual-heap* 1)
      (setq heapupperbound (wgetv *the-heapupperbounds* *first-heap*)) 
      (setq heaptrapbound (wplus2 heapupperbound -120))
      (copyd '!%reclaim '!%rreclaim)
))

(de propagate-1-heap (x)
     (if (wlessp *number-of-heaps* *max-number-of-heaps*)
          (if (eq x *number-of-heaps*) (setq x (wplus2 x -1))
                                       (setq x (wplus2 x 1)))
  % as long as not all heap chunks are filled just use 2 all the
  % time to save paging overhead
      (setq x (iadd1 x))
      (when (wgreaterp  x *number-of-heaps*) (setq x 1)))
      x)
 
(de propagate-heaps ()
      (wputv *the-heaplasts* *actual-heap*  heaplast) % save heaplast  
      (setq *next-heap*       (propagate-1-heap *next-heap*))
      (setq *actual-heap*     (propagate-1-heap *actual-heap*))
      (setq heaplast          (wgetv *the-heaplowerbounds* *actual-heap*))
      (setq heaplowerbound    (wgetv *the-heaplowerbounds* *actual-heap*))
      (setq heapupperbound    (wgetv *the-heapupperbounds* *actual-heap*))
      (setq oldheapupperbound (wgetv *the-heapupperbounds* *first-heap*))
      (setq oldheaplowerbound (wgetv *the-heaplowerbounds* *first-heap*))
      (setq oldheaplast       (wgetv *the-heaplasts* *first-heap*))
      (setq heaptrapbound     (wdifference heapupperbound 120))
      (if (eq *next-heap* *first-heap*) nil t)
)

(de !%rreclaim ()
  (prog (b oldtimehi oldtimelo oldfirst) 
  (setq *bruch* 17)
  (setq oldfirst *first-heap*)
  (when *gc  (errorprintf "*** BIGGC Garbage collection starting"))
  (beforegcsystemhook)

  (when (propagate-heaps) (return nil)) % no reclaim necessary
 
  (setq *grosso-gc* NIL)  % mark all id's 
  (setq oldtimehi (timc)) % a strange problem after 37 hours
  (setq oldtimelo (wand oldtimehi 16#ffff))
  (setq oldtimehi (wshift oldtimehi -16))
  (setq oldsize (wdifference oldheaplast oldheaplowerbound))
  (setf gcknt* (+ gcknt* 1))

  (copyfromallbases)

  (wputv *the-heaplasts* oldfirst (wgetv *the-heaplowerbounds* oldfirst))
       % set old heap as empty for next round
  (for (from i 1 *number-of-heaps*)  (do (copyfromonestaticheap i)))
  (setq *first-heap* (propagate-1-heap *first-heap*))

  (makeidfreelist)
  (aftergcsystemhook)

  (setq oldtime (wor (wshift oldtimehi 16) oldtimelo))
  (setq oldtime (wdifference (timc) oldtime))
  (setf gctime* (wplus2 gctime* oldtime))
  (when (not (posintp gctime*)) (setq gctime* (inf -1)))
                                     % another 37 hours problem
  (when *gc  (gcstats))
  (setq heaptrapped nil)
  (when (wlessp *bruch* 0) (lisprpv))
  (setq *bruch* 0)

 % feature for enlarging the heap-space:
 % heap is separated into many chunks most of which are "closed" in the
 % beginning to keep paging small as long as not many items are really
 % stable. If we are full, in this case less than 200000 cells are free
 % another chunk is opened and we continue with the new and the last one
  (when (and (wlessp (intinf (known-free-space)) 200000)
             (eq  *next-heap* *number-of-heaps*)
             (not (eq *number-of-heaps* *max-number-of-heaps*)))
              (prin2t "+++ Opening additional heap space +++")
              (wputv *the-heaplasts* oldfirst
                       (wgetv *the-heaplowerbounds* oldfirst))
              (wputv *the-heaplasts* *first-heap* heaplast)
            % mark old first heap as invalid (full of forwards)
              (setq *actual-heap* *next-heap*)
              (setq *first-heap* *next-heap*)
              (setq *next-heap* (iadd1 *number-of-heaps*))
              (setq heaplast (wgetv *the-heaplowerbounds* *actual-heap*))
              (setq heaptrapbound
                     (wplus2 (wgetv *the-heapupperbounds* *actual-heap*) -120))
              (setq heaplowerbound (wgetv *the-heaplowerbounds* *actual-heap*))
              (setq heapupperbound (wgetv *the-heapupperbounds* *actual-heap*)) 
              (setq *number-of-heaps* (iadd1 *number-of-heaps*))
              (when (eq *number-of-heaps* *max-number-of-heaps*)
                    (setq *first-heap* 1)))
               % changing to circular mode
  (when (wlessp (intinf (known-free-space)) (intinf heap-warn-level))
    (continuableerror 99 "Heap space low" nil))

  nil
  ))


%(commentoutcode
(de CopyFromOneStaticHeap (i)
 (prog (memo B upperb Size memo1 tagmemo
        localoldheaplowerbound localoldheaplast) 

    (setq localoldheaplowerbound oldheaplowerbound)
    (setq localoldheaplast oldheaplast)

    (when (or (eq i *first-heap*) (eq i *actual-heap*)) (return nil))

    (setq B (wgetv *the-heaplowerbounds* i)) %%%%StaticHeapBase) 
    (setq upperb (wgetv *the-heaplasts* i))
    (unixputn i) (console-print-string " 1 static ")
    (unixputn b) (console-print-string " 1 static ")
    (unixputn upperb) (console-newline)
    (while (wlessp B upperb) 
       (setq memo (getmem b))
       (case (Tag memo)
        ((hwords-tag) 
              (progn  (setq Size (ashift (wshift memo 5) -5)) % signed inf
                      (setq B (wplus2 B (<< Size 2) 8))))
        ((hbytes-tag) 
              (progn  (setq Size (ashift (wshift memo 5) -5)) % signed inf
                      (setq B (wplus2 B (<< (>> (wplus2 Size 5) 2) 2) 4)))) 
        ((hhalfwords-tag) 
              (progn  (setq Size (ashift (wshift memo 5) -5)) % signed inf
                      (setq B (wplus2 B (<< (>> (wplus2 Size 2) 1) 2) 4))))
        ((hvect-tag)
              (progn  (setq Size (ashift (wshift memo 5) -5)) % signed inf
                      (setq B (wplus2 B 4))  % skip header word
                      (for (from N 0 Size 1) (do 
                             (setq memo1 (getmem b))
                             (setq tagmemo (tag memo1))
                             (when (and (pointertagp tagmemo)
                                        (withinlocaloldheappointer (inf memo1)))
                                             (CopyItem B memo1)) 
                             (setq B (wplus2 B 4))))))
        ((0 11 12 13 14 15 16 17 18 19 20 21 22 27 28 29 30 31)
                  (setq B (wplus2 B 4))) % not pointertag
        ((1 2 3 4 5 6 7 8 9 10 )
             (progn     % pairs
	      (when (not (eq 0 (wand 3 (inf memo))))
		(centerdump b 200) (centerdump (inf memo) 200)
		(console-print-string " BUS kaputt") (quit))
              (when (withinlocaloldheappointer (inf memo))
                                           (copyfromonestackitem b)) 
              (setq B (wplus2 B 4))))
)))
) %end commentoutcode


(lap '(
(*ENTRY COPYFFROMONESTATICHEAP EXPR 1)
(*ALLOC 12)
(*MOVE (REG 1) (FRAME 1))
(*MOVE (QUOTE NIL) (FRAME 2))
(*MOVE (QUOTE NIL) (FRAME 3))
(*MOVE (QUOTE NIL) (FRAME 4))
(*MOVE (QUOTE NIL) (FRAME 5))
(*MOVE (QUOTE NIL) (FRAME 6))
(*MOVE (QUOTE NIL) (FRAME 7))
(*MOVE (QUOTE NIL) (FRAME 8))
(*MOVE ($GLOBAL OLDHEAPLOWERBOUND) (FRAME 9))
(*MOVE ($GLOBAL OLDHEAPLAST) (FRAME 10))
(*JUMPEQ (LABEL G0006) (REG 1) ($FLUID *FIRST-HEAP*))
(*JUMPNOTEQ (LABEL G0004) (REG 1) ($FLUID *ACTUAL-HEAP*))
(*JUMP (LABEL G0008))
(*LBL (LABEL G0006))
(*MOVE (QUOTE NIL) (REG 1))
(*EXIT 12)
(*LBL (LABEL G0004))
(*MOVE (FRAME 1) (REG 1))
(*WSHIFT (REG 1) (WCONST 2))
(*WPLUS2 (REG 1) ($FLUID *THE-HEAPLOWERBOUNDS*))
(*MOVE (MEMORY (REG 1) (WCONST 0)) (FRAME 3))
(*MOVE (FRAME 1) (REG 2))
(*WSHIFT (REG 2) (WCONST 2))
(*WPLUS2 (REG 2) ($FLUID *THE-HEAPLASTS*))
(*MOVE (MEMORY (REG 2) (WCONST 0)) (FRAME 4))
        (MOVE-RELOCATABLE-OBJECT "L0061" (REG l7))
(*LBL (LABEL G0009))
(*MOVE (MEMORY (FRAME 3) (WCONST 0)) (FRAME 2))
%%%%%%%% (*JUMPWGEQ (LABEL G0008) (FRAME 3) (FRAME 4))
	(SUBCC (REG L3) (REG L4) (REG G0))
	(BGE (LABEL G0008))
(*FIELD (REG 1) (FRAME 2) (QUOTE 0) (QUOTE 5))
	(SLL (REG 1) 3 (REG T5))
	(JMPL (DISPLACEMENT (REG l7) (REG T5)) (REG 0))
	(or (reg g0) 0 (reg g0))
     "L0061"
	(BA (LABEL G0009))
	(add (reg l3) 4 (reg l3))
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
	(BA (LABEL G0019))
	(NOP)
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3)) 
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3)) 
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3)) 
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3)) 
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3)) 
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3)) 
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
	(BA (LABEL G0015))
	(NOP)
	(BA (LABEL G0016))
	(NOP)
	(BA (LABEL G0014))
	(NOP)
	(BA (LABEL G0017))
	(NOP)
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
        (BA (LABEL G0009))
        (add (reg l3) 4 (reg l3))
(*JUMP (LABEL G0009))
(*LBL (LABEL G0014))
(*MOVE (FRAME 2) (REG 1))
(*WSHIFT (REG 1) (QUOTE 5))
(*ASHIFT (REG 1) (QUOTE -5))
(*MOVE (REG 1) (FRAME 5))
(*MOVE (WCONST 8) (REG 2))
(*WPLUS2 (REG 2) (FRAME 3))
(*MOVE (FRAME 5) (REG 3))
(*WSHIFT (REG 3) (QUOTE 2))
(*WPLUS2 (REG 2) (REG 3))
(*MOVE (REG 2) (FRAME 3))
(*JUMP (LABEL G0009))
(*LBL (LABEL G0015))
(*MOVE (FRAME 2) (REG 1))
(*WSHIFT (REG 1) (QUOTE 5))
(*ASHIFT (REG 1) (QUOTE -5))
(*MOVE (REG 1) (FRAME 5))
(*MOVE (WCONST 4) (REG 2))
(*WPLUS2 (REG 2) (FRAME 3))
(*MOVE (WCONST 5) (REG 3))
(*WPLUS2 (REG 3) (FRAME 5))
(*WSHIFT (REG 3) (WCONST -2))
(*WSHIFT (REG 3) (QUOTE 2))
(*WPLUS2 (REG 2) (REG 3))
(*MOVE (REG 2) (FRAME 3))
(*JUMP (LABEL G0009))
(*LBL (LABEL G0016))
(*MOVE (FRAME 2) (REG 1))
(*WSHIFT (REG 1) (QUOTE 5))
(*ASHIFT (REG 1) (QUOTE -5))
(*MOVE (REG 1) (FRAME 5))
(*MOVE (WCONST 4) (REG 2))
(*WPLUS2 (REG 2) (FRAME 3))
(*MOVE (WCONST 2) (REG 3))
(*WPLUS2 (REG 3) (FRAME 5))
(*WSHIFT (REG 3) (WCONST -1))
(*WSHIFT (REG 3) (QUOTE 2))
(*WPLUS2 (REG 2) (REG 3))
(*MOVE (REG 2) (FRAME 3))
(*JUMP (LABEL G0009))
(*LBL (LABEL G0017))
(*MOVE (FRAME 2) (REG 1))
(*WSHIFT (REG 1) (QUOTE 5))
(*ASHIFT (REG 1) (QUOTE -5))
(*MOVE (REG 1) (FRAME 5))
(*WPLUS2 (FRAME 3) (WCONST 4))
(*MOVE (FRAME 5) (REG 2))
(*MOVE (QUOTE 0) (REG 1))
(*MOVE (REG 1) (FRAME 11))
(*MOVE (REG 2) (FRAME 12))
(*LBL (LABEL G0025))
(*JUMPWGREATERP (LABEL G0009) (FRAME 11) (FRAME 12))
(*MOVE (MEMORY (FRAME 3) (WCONST 0)) (FRAME 6))
(*FIELD (REG 1) (FRAME 6) (QUOTE 0) (QUOTE 5))
(*MOVE (REG 1) (FRAME 8))
(*MOVE (WCONST -1) (REG 2))
(*WPLUS2 (REG 2) (FRAME 8))
(*MOVE (WCONST 19) (REG 3))
(*WDIFFERENCE (REG 3) (FRAME 8))
(*WOR (REG 2) (REG 3))
(*JUMPWLESSP (LABEL G0029) (REG 2) (QUOTE 0))
(*FIELD (REG 4) (FRAME 6) (QUOTE 5) (QUOTE 27))
(*WDIFFERENCE (REG 4) (FRAME 9))
(*FIELD (REG 5) (FRAME 6) (QUOTE 5) (QUOTE 27))
(*WMINUS (REG 5) (REG 5))
(*WPLUS2 (REG 5) (FRAME 10))
(*WOR (REG 4) (REG 5))
(*JUMPWLESSP (LABEL G0029) (REG 4) (QUOTE 0))
(*MOVE (FRAME 3) (REG 1))
(*move (indexed (reg 1) 0) (reg 2))
(*LINK COPYITEM EXPR 2)
(*LBL (LABEL G0029))
(*WPLUS2 (FRAME 3) (WCONST 4))
(*WPLUS2 (FRAME 11) (WCONST 1))
(*JUMP (LABEL G0025))
(*LBL (LABEL G0018))
(*WPLUS2 (FRAME 3) (WCONST 4))
(*JUMP (LABEL G0009))
(*LBL (LABEL G0019))
(*FIELD (REG 1) (FRAME 2) (QUOTE 5) (QUOTE 27))
(*WDIFFERENCE (REG 1) (FRAME 9))
(*FIELD (REG 2) (FRAME 2) (QUOTE 5) (QUOTE 27))
%(*WMINUS (REG 2) (REG 2))
%(*WPLUS2 (REG 2) (FRAME 10))
(sub (reg i2) (REG 2) (REG 2))
(*WOR (REG 1) (REG 2))
(*JUMPWLESSP (LABEL G0033) (REG 1) (QUOTE 0))
(*MOVE (FRAME 3) (REG 1))
(*move (indexed (reg 1) 0) (reg 2)) 
(*LINK COPYITEM EXPR 2)
(*LBL (LABEL G0033))
(*WPLUS2 (FRAME 3) (WCONST 4))
(*JUMP (LABEL G0009))
(*LBL (LABEL G0008))
(*MOVE (QUOTE NIL) (REG 1))
(*EXIT 12)
))

(lap '(
(*ENTRY COPYITEM EXPR 2)
(*LBL (LABEL G0002))
(*JUMPNOTEQ (LABEL G0001) (REG 2) (QUOTE NIL))
(*LBL (LABEL G0008))   % ints etc
(jmpl (indexed (reg o7) 8) (reg g0))
(*move (reg g6) (reg 1))
(*lbl (label g0001))
(*FIELD (REG 5) (REG 2) (QUOTE 0) (QUOTE 5))
(*JUMPON (REG 5) 0 31 (
  (LABEL G0008) (LABEL G0015) (LABEL G0016) (LABEL G0013) (LABEL G0012)
  (LABEL G0011) (LABEL G0017) (LABEL G0015) (LABEL G0014) (LABEL G0010)
  (LABEL G0014) (LABEL G0008) (LABEL G0008) (LABEL G0008) (LABEL G0008)
  (LABEL G0018) (LABEL G0018) (LABEL G0018) (LABEL G0018) (LABEL G0018)
  (LABEL G0008) (LABEL G0018) (LABEL G0008) (LABEL G0018) (LABEL G0018)
  (LABEL G0007) (LABEL G0018) (LABEL G0018) (LABEL G0008) (LABEL G0008)
  (LABEL G0009) (LABEL G0008)))
(*JUMP (LABEL G0166))

(*LBL (LABEL G0009)) % id case
(*move (fluid SYMNAM) (reg 5))
(*FIELD (reg 1) (reg 2) (quote 5) (quote 27))
(*MOVE (REG 1) (reg 4))
(*WSHIFT (REG 4) (WCONST 2))
(*WPLUS2 (REG 4) (reg 5))
(*FIELD (REG 4) (MEMORY (REG 4) (WCONST 0)) (QUOTE 0) (QUOTE 5))
(*JUMPEQ (LABEL G0008) (REG 4) (QUOTE 27))
(*jcall MARKANDCOPYFROMID EXPR 1)

(*LBL (LABEL G0010))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27))
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0023) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0023))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0031) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0031))
(*MOVE ($GLOBAL HEAPLAST) (REG 1))
(*MOVE (REG 1) (FRAME 7))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (CAR (REG 2)))
(*WPLUS2 ($GLOBAL HEAPLAST) (WCONST 8))
(*MOVE (FRAME 7) (REG 1))
(*MKITEM (REG 1) (QUOTE 9))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*MOVE (FRAME 5) (MEMORY (FRAME 7) (WCONST 0)))
(*MOVE (CDR (REG 2)) (REG 1))
(*MOVE (REG 1) (MEMORY (FRAME 7) (WCONST 4)))
(*MOVE (MEMORY (FRAME 7) (WCONST 0)) (REG 2))
(*MOVE (WCONST 0) (REG 1))
(*WPLUS2 (REG 1) (FRAME 7))
(*LINK COPYITEM EXPR 2)
(*MOVE (MEMORY (FRAME 7) (WCONST 4)) (REG 2))
(*MOVE (WCONST 4) (REG 1))
(*WPLUS2 (REG 1) (FRAME 7))
(*dealloc 11)
(*JUMP (LABEL G0002))

(*LBL (LABEL G0011))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27)) 
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0048) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0048))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0051) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0051))
(*MOVE (REG 2) (REG 1))
(*LINK COPYBYTES EXPR 1)
(*MOVE (REG 1) (FRAME 7))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (MEMORY (FRAME 4) (WCONST 0)))
(*MOVE (FRAME 7) (REG 1))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)

(*LBL (LABEL G0012))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27)) 
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0055) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0055))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0058) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0058))
(*MOVE (REG 2) (REG 1))
(*LINK COPYSTRING EXPR 1)
(*MOVE (REG 1) (FRAME 7))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (MEMORY (FRAME 4) (WCONST 0)))
(*MOVE (FRAME 7) (REG 1))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)

(*LBL (LABEL G0013))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27)) 
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0062) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0062))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0065) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0065))
(*LINK GTFLTN EXPR 0)
(*MOVE (REG 1) (FRAME 7))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (MEMORY (FRAME 4) (WCONST 0)))
(*MOVE (MEMORY (FRAME 4) (WCONST 4)) (MEMORY (FRAME 7) (WCONST 4)))
(*MOVE (MEMORY (FRAME 4) (WCONST 8)) (MEMORY (FRAME 7) (WCONST 8)))
(*MOVE (FRAME 7) (REG 1))
(*MKITEM (REG 1) (QUOTE 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)

(*LBL (LABEL G0014))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27)) 
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0069) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0069))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0072) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0072))
(*MOVE (FRAME 4) (FRAME 6))
(*SIGNEDFIELD (REG 1) (MEMORY (FRAME 6) (WCONST 0)) (QUOTE 5) (QUOTE 27))
(*MOVE (REG 1) (FRAME 9))
(*MOVE (FRAME 9) (REG 1))
(*LINK GTVECT EXPR 1)
(*MOVE (REG 1) (FRAME 8))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (MEMORY (FRAME 6) (WCONST 0)))
(*MOVE (FRAME 9) (REG 2))
(*MOVE (QUOTE 0) (REG 1))
(*MOVE (REG 1) (FRAME 10))
(*MOVE (REG 2) (FRAME 11))
(*LBL (LABEL G0141))
(*JUMPWGREATERP (LABEL G0140) (FRAME 10) (FRAME 11))
(*MOVE (WCONST 1) (REG 1))
(*WPLUS2 (REG 1) (FRAME 10))
(*WSHIFT (REG 1) (WCONST 2))
(*WPLUS2 (REG 1) (FRAME 8))
(*MOVE (WCONST 1) (REG 2))
(*WPLUS2 (REG 2) (FRAME 10))
(*WSHIFT (REG 2) (WCONST 2))
(*WPLUS2 (REG 2) (FRAME 6))
(*MOVE (MEMORY (REG 2) (WCONST 0)) (MEMORY (REG 1) (WCONST 0)))
(*MOVE (WCONST 1) (REG 2))
(*WPLUS2 (REG 2) (FRAME 10))
(*WSHIFT (REG 2) (WCONST 2))
(*WPLUS2 (REG 2) (FRAME 8))
(*MOVE (MEMORY (REG 2) (WCONST 0)) (REG 2))
(*MOVE (WCONST 1) (REG 1))
(*WPLUS2 (REG 1) (FRAME 10))
(*WSHIFT (REG 1) (WCONST 2))
(*WPLUS2 (REG 1) (FRAME 8))
(*LINK COPYITEM EXPR 2)
(*WPLUS2 (FRAME 10) (WCONST 1))
(*JUMP (LABEL G0141))
(*LBL (LABEL G0140))
(*MOVE (FRAME 8) (REG 1))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)

(*LBL (LABEL G0015))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27)) 
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0145) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0145))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0148) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0148))
(*MOVE (REG 2) (REG 1))
(*LINK COPYWRDS EXPR 1)
(*MOVE (REG 1) (FRAME 7))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (MEMORY (FRAME 4) (WCONST 0)))
(*MOVE (FRAME 7) (REG 1))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)

(*LBL (LABEL G0016))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27)) 
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0152) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0152))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0155) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0155))
(*MOVE (REG 2) (REG 1))
(*LINK COPYWRDS EXPR 1)
(*MOVE (REG 1) (FRAME 7))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (MEMORY (FRAME 4) (WCONST 0)))
(*MOVE (FRAME 7) (REG 1))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)

(*LBL (LABEL G0017))
(*FIELD (reg 3) (reg 2) (quote 5) (quote 27)) 
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (REG 3) (FRAME 4))
(*MOVE (FRAME 4) (REG 4))
(*WDIFFERENCE (REG 4) ($GLOBAL OLDHEAPLOWERBOUND))
(*MOVE ($GLOBAL OLDHEAPLAST) (REG 5))
(*WDIFFERENCE (REG 5) (FRAME 4))
(*WOR (REG 4) (REG 5))
(*JUMPWGEQ (LABEL G0159) (REG 4) (QUOTE 0))
(*MOVE (REG 2) (REG 1))
(*EXIT 11)
(*LBL (LABEL G0159))
(*MOVE (MEMORY (FRAME 4) (WCONST 0)) (FRAME 5))
(*FIELD (REG 1) (FRAME 5) (QUOTE 0) (QUOTE 5))
(*JUMPNOTEQ (LABEL G0162) (REG 1) (QUOTE 27))
(*FIELD (REG 1) (FRAME 5) (QUOTE 5) (QUOTE 27))
(*MKITEM (REG 1) (FRAME 3))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)
(*LBL (LABEL G0162))
(*MOVE (REG 2) (REG 1))
(*LINK COPYHALFWORDS EXPR 1)
(*MOVE (REG 1) (FRAME 7))
(*MKITEM (REG 1) (QUOTE 27))
(*MOVE (REG 1) (MEMORY (FRAME 4) (WCONST 0)))
(*MOVE (FRAME 7) (REG 1))
(*MKITEM (REG 1) (QUOTE 6))
(*MOVE (REG 1) (MEMORY (FRAME 1) (WCONST 0)))
(*EXIT 11)

(*LBL (LABEL G0018))
(*FIELD (REG 4) (REG 2) (QUOTE 5) (QUOTE 27))
(*ALLOC 11) 
(*move (reg 5) (frame 3))
(*MOVE (REG 1) (FRAME 1))
(*MOVE (REG 2) (FRAME 2))
(*MOVE (FRAME 3) (REG 3))
(*MOVE (FRAME 1) (REG 2))
(*MOVE (QUOTE "*** In %w: Unexpected tag %w found at %w during gc %n") (REG 1))
(*LINK ERRORPRINTF EXPR 4)
(*LBL (LABEL G0166))
(*LBL (LABEL G0007))
(*MOVE (QUOTE NIL) (REG 1))
(*EXIT 11)
))

(compiletime (off fast-integers))
