%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           CRAY-specifics
% Title:          Run-time support for assorted things (mostly in assembler
%                 or vector code)
% Author:         H. Melenk and W. Neun
% Created:        26 June 1986
% Status:         Experimental
% Mode:           Lisp
% Package:        Nonkernel
%
% (c) Copyright 1987, Konrad Zuse-Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 20-Nov-1987 W. Neun
% Discovered that DRI/ERI instructions wont run on a Cray -1 model
% Therefore made DRI/ERI dependent on if_system CRAY1
 
% 20-Mar-1987 W. Neun
% changed most functions to use the DRI/ERI instructions.
% (ENABLE/DISABLE operand range interrupts) to do preloads!
 
(compiletime (progn
              (define-constant try-string-length 6)
              (Load vector-instructions if-system)))
 
%(fluid '(symcnt symprp symval   % all fluids in Regs must be known now
%         &symflags&  symget ))  % because their values will be unbound
                                % at the bottom. A fluid ... later
                                % will set values to nil!!!
 
(copyd 'old-unchecked-string-equal 'unchecked-string-equal)
 
(remd 'unchecked-string-equal) %be quiet when redefined
 
(de unchecked-string-equal (u v)
   (prog (len1 restlen1 lenw1 shiftsize)
       (VSetVl try-string-length)
       (VSet (VReg 0) -1)
       (VLoad (VReg 6) u 1)
       (VSetelement (VReg 0) 0 (WRshift -1 27)) %ignore tag and gc field
       (VandV (VReg 3) (VReg 6) (VReg 0))
       (VLoad (VReg 2) v 1)
       (VandV (VReg 4) (VReg 2) (VReg 0))
       (setq len1 (VGetelement (VReg 3) 0))
       (setq lenw1 (strpack len1))
       (if (wgeq lenw1 try-string-length)
           (progn
              (when (wgreaterp len1 (wlshift 1 30))  % case -1 ("")
                    (return (eq len1 (vgetelement (Vreg 4) 0))))
              (VminusV (VReg 5) (VReg 4) (VReg 3))
              (when (wneq (Vtestn (VReg 5)) 0) (return nil))
              (return (old-unchecked-string-equal u v)))
           (Progn
              (VsetVl Lenw1)
              (VminusV (VReg 5) (VReg 4) (VReg 3))
              (when (wneq (Vtestn (VReg 5)) 0) (return nil))
              (setq restlen1 (wand (wplus2 len1 1) 7))
              (when (weq restlen1 0) (return t))
              (setq shiftsize (wlshift (wdifference 8 restlen1) 3))
              (return
                 (eq (wRshift (VGetElement (VReg 6) Lenw1) shiftsize)
                     (wRshift (VGetElement (VReg 2) Lenw1) shiftsize))
)   )  )) )
 
(compiletime (off lapopt))
 
(if_system CRAY1
 
 (lap '((*entry !%get expr 2)  % this get is about 50 % faster
       (rmask (Reg s7) 14)     % for the code see the lisp version
       (*wand (Reg s7) (Reg 1))  % and Reg1 = pl
                                 % and Reg2 = v
                                 % and Reg3 = x
                                 % and Reg4 = previouspointer
                                 % and Reg5 for fetch ahead
       (*Move (Reg symprp) (Reg a6))
       (*Move (Reg s7) (Reg a7))
       (*wplus2 (Reg a6) (Reg a7))
       (*wshift (Reg 1) -59)
       (*Move (Indexed (Reg a6) 0) (Reg 5))
       (*Move id-tag (Reg s7))
       (*Wxor3 (Reg s0) (Reg 1) (Reg s7))
       (*Move (Reg 2) (Reg s6))
       (*wshift (Reg s6) -59)
       (*Wxor3 (Reg s6) (Reg s6) (Reg s7))
       (JSN (label returnnil))        %last statment of id-test
       (*Move (Reg s6) (Reg s0))
       (*Move (Reg 5) (Reg a6)) % 2
       (*Move (Indexed (Reg a6) 0) (Reg 3)) % moved from below
       (JSN (Label returnnil))
 
       (*Move (Reg 5) (Reg 1))
       (*Move (Reg t00) (Reg 5))
       (*Wxor3 (Reg s0) (Reg 1) (Reg 5))
       (*Move (Indexed (Reg a6) 1) (Reg 5))
       (*Move 9 (Reg s7))              %pair
       (*Move (Reg 3) (Reg s6))
       (*Move (Reg 3) (Reg T71))
       (*wshift (Reg s6) -59)
       (JSZ (label returnnil))
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7))
       (*Move (Reg 3) (Reg a6))
       (*Move (Reg 1) (Reg 4))
       (JSN (label nopair))
       (*Move (Indexed (Reg a6) 0) (Reg s6))
       (*Move (Indexed (Reg a6) 1) (Reg s7))
       (*Wxor3 (Reg s0) (Reg s6) (Reg 2))
       (*Move (Reg 5) (Reg a7))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
       (JSN (label loop))
       (*Move (Reg s7) (Reg 1))
       (j (Reg b00))
     returnnil
       (*Move (Reg T00) (Reg 1))
       (j (Reg b00))
     nopair
       (*Move (Reg 5) (Reg a7))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
     loop
       (*Move 9 (Reg s7))              %pair
       (*Move (Reg 5) (Reg 1))
       (*Move (Reg t00) (Reg 5))
       (*Wxor3 (Reg s0) (Reg 1) (Reg 5))
       (*Move (Indexed (Reg a7) 1) (Reg 5))
       (*Move (Reg 3) (Reg s6)) % 1 from below
       (JSZ (label returnnil))
       (*wshift (Reg s6) -59)
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7))
       (*Move (Reg 3) (Reg a6))
       (JSN (label nopair1))
       (*Move (Indexed (Reg a6) 0) (Reg s6))
       (*Move (Indexed (Reg a6) 1) (Reg s7))
       (*Move (Reg 5) (Reg a7))
       (*Wxor3 (Reg s0) (Reg s6) (Reg 2))
       (JSZ (label retcdr1))
       (*Move (Reg 3) (Reg T71))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
       (*Move (Reg 1) (Reg 4))
       (J (label loop))
     nopair1
       (*Move (Reg 5) (Reg a7))
       (*Move (Reg 3) (Reg T71))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
       (*Move (Reg 1) (Reg 4))
       (J (label loop))
     retcdr1
       (*Move (Reg 4) (Reg a6))
       (*Move (Reg T71) (Reg s6))
       (*Move (Reg 1) (Reg a7))
       (*Move (Reg s7) (Reg 1))
       (*Move (Reg 3) (Indexed (Reg a6) 0))
       (*Move (Reg s6) (Indexed (Reg a7) 0))
       (j (Reg b00))
)     )
 (lap '((*entry !%get expr 2)  % this get is about 50 % faster
       (dri)  % for the code see the lisp version
                                 % and Reg1 = pl
                                 % and Reg2 = v
                                 % and Reg3 = x
                                 % and Reg4 = previouspointer
                                 % and Reg5 for fetch ahead
       (*Move (Reg symprp) (Reg a6))
       (*Move (Reg 1) (Reg a7))
       (*wplus2 (Reg a6) (Reg a7))
       (*wshift (Reg 1) -59)
       (*Move (Indexed (Reg a6) 0) (Reg 5))
       (*Move id-tag (Reg s7))
       (*Wxor3 (Reg s0) (Reg 1) (Reg s7))
       (*Move (Reg 2) (Reg s6))
       (*wshift (Reg s6) -59)
       (*Wxor3 (Reg s6) (Reg s6) (Reg s7))
       (JSN (label returnnil))        %last statment of id-test
       (*Move (Reg s6) (Reg s0))
       (*Move (Reg 5) (Reg a6)) % 2
       (*Move (Indexed (Reg a6) 0) (Reg 3)) % moved from below
       (JSN (Label returnnil))
 
       (*Move (Reg 5) (Reg 1))
       (*Move (Reg t00) (Reg 5))
       (*Wxor3 (Reg s0) (Reg 1) (Reg 5))
       (*Move (Indexed (Reg a6) 1) (Reg 5))
       (*Move 9 (Reg s7))              %pair
       (*Move (Reg 3) (Reg s6))
       (*Move (Reg 3) (Reg T71))
       (*wshift (Reg s6) -59)
       (JSZ (label returnnil))
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7))
       (*Move (Reg 3) (Reg a6))
       (dri)       %%%%%%%%%%%%%%%%%%%%%%%%%%
       (*Move (Indexed (Reg a6) 0) (Reg s6))
       (*Move (Indexed (Reg a6) 1) (Reg s7))
       (*Move (Reg 1) (Reg 4))
       (JSN (label nopair))
       (*Wxor3 (Reg s0) (Reg s6) (Reg 2))
       (*Move (Reg 5) (Reg a7))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
       (JSN (label loop))
       (*Move (Reg s7) (Reg 1))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       (j (Reg b00))
     returnnil
       (*Move (Reg T00) (Reg 1))
       (*move (Reg 5) (Reg 2))    % synchronize
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       (j (Reg b00))
     nopair
       (*Move (Reg 5) (Reg a7))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
     loop
       (*Move 9 (Reg s7))              %pair
       (*Move (Reg 5) (Reg 1))
       (*Move (Reg t00) (Reg 5))
       (*Wxor3 (Reg s0) (Reg 1) (Reg 5))
       (*Move (Indexed (Reg a7) 1) (Reg 5))
       (*Move (Reg 3) (Reg s6)) % 1 from below
       (JSZ (label returnnil))
       (*wshift (Reg s6) -59)
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7))
       (*Move (Reg 3) (Reg a6))
       (*Move (Indexed (Reg a6) 0) (Reg s6)) % 2 from below
       (*Move (Indexed (Reg a6) 1) (Reg s7)) % using DRI
       (JSN (label nopair1))
       (*Move (Reg 5) (Reg a7))
       (*Wxor3 (Reg s0) (Reg s6) (Reg 2))
       (JSZ (label retcdr1))
       (*Move (Reg 3) (Reg T71))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
       (*Move (Reg 1) (Reg 4))
       (J (label loop))
     nopair1
       (*Move (Reg 5) (Reg a7))
       (*Move (Reg 3) (Reg T71))
       (*Move (Indexed (Reg a7) 0) (Reg 3))
       (*Move (Reg 1) (Reg 4))
       (J (label loop))
     retcdr1
       (*Move (Reg 4) (Reg a6))
       (*Move (Reg T71) (Reg s6))
       (*Move (Reg 1) (Reg a7))
       (*Move (Reg s7) (Reg 1))
       (*Move (Reg 3) (Indexed (Reg a6) 0))
       (*Move (Reg s6) (Indexed (Reg a7) 0))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       (j (Reg b00))
))     ) 
 
(compiletime (flag '(hard-put-symfnc atsoc-intern)'internalfunction))
 
(de hard-put-symfnc()
   (putmem (wplus2 symfnc (id2int 'get))
                (getmem (wplus2 symfnc (id2int '!%get)))) nil)
 
(hard-put-symfnc)
 
(de start-of-equal (x) x)
%
% The following version of EQUAL is alloc-free for all but the pair
% case, in which the alloc is done in pairequal.
% functions pairequal and ecci must be close together in memory
% to hopefully reside in the same instruction buffer
%
(compiletime (flag '(pairequal ecci) 'internalfunction))
 
(de pairequal(u v w x)
   (prog (tag1 tag2)
     (setq tag1 (tag u))
     (setq tag2 (tag v))
     (if (or (eq u v) (and (not (wgeq tag2 id-tag)) %does negint also
                           (not (eq tag2 posint-tag))
                           (eq tag1 tag2) (ecci u v)))
             (progn
                (setq tag1 (tag w))
                (setq tag2 (tag x))
                (cond  ((eq w x) (return t))
                       ((not (eq tag1 tag2)) (return nil))
                       ((eq tag1 pair-tag)
                             (return (pairequal (car x) (car w)
                                                (cdr x) (cdr w))))
                       ((wgeq tag1 id-tag) (return nil)) %negint also
                       ((eq tag2 posint-tag) (return nil))
                       (t (return (ecci w x)))
              ) )
              (return nil))
) )
 
(compiletime
 (progn
   (ds nontrivial (form)    (and (eq 0 (wor (wdifference caru carv)
                                            (wdifference cdru cdrv)))
                                  form))
   (ds nontrivialstring (form)  (and (eq caru carv) form))
))
 
(if_system CRAY1
 (de ecci  (u v)
  % Structural equality,internalfunction version of equal
  % need not check types again; eq case already checked
  % must not expect ints or ids here, but use "fast" jumpon (see CMACS3)
 
    (if (pairp u) (pairequal (car u) (car v) (cdr u) (cdr v))
     (case (tag u)
      ((posint-tag negint-tag) (eq u v))
      ((evector-tag)   (evectorequal u v))
      ((vector-tag)    (vectorequal u v))
      ((string-tag)    (unchecked-string-equal u v))
      ((bytes-tag)     (unchecked-string-equal u v))
      ((floatnum-tag)  (weq (cdr u) (cdr v)))  % not nice but it fits
      ((fixnum-tag)    (weq (cdr u) (cdr v)))
      ((bignum-tag)    (wordsequal u v))
      ((words-tag)     (wordsequal u v))
      ((halfwords-tag) (halfwordsequal u v))
      ((code-tag hbytes-tag forward-tag) nil)   %for fast jumpon
      ((pair-tag) nil)
      (nil nil))))
 
(de ecci  (u v)    % Cray X-MP
  % Structural equality,internalfunction version of equal
  % need not check types again; eq case already checked
  % must not expect ints or ids here, but use "fast" jumpon (see CMACS3)
 
   (disable-operand-range-interrupt)
   %%%% !!! %%%%%%%
   (let  ((caru (car u)) (cdru (cdr u)) (carv (car v)) (cdrv (cdr v)))
 
     (if (pairp u) (progn (enable-operand-range-interrupt)
                          (pairequal caru carv cdru cdrv))
     (setq caru (inf caru))
     (setq carv (inf carv))
     (enable-operand-range-interrupt)
     (case (tag u)
      ((posint-tag negint-tag) (eq u v))
      ((evector-tag)   (nontrivial (evectorequal u v)))
      ((vector-tag)    (nontrivial (vectorequal u v)))
      ((string-tag)    (nontrivialstring (unchecked-string-equal u v)))
      ((bytes-tag)     (nontrivialstring (unchecked-string-equal u v)))
      ((floatnum-tag)  (weq cdru cdrv))
      ((fixnum-tag)    (weq cdru cdrv))
      ((bignum-tag)    (nontrivial (wordsequal u v)))
      ((words-tag)     (nontrivial (wordsequal u v)))
      ((halfwords-tag) (halfwordsequal u v))
      ((code-tag hbytes-tag forward-tag) nil)   %for fast jumpon
      ((pair-tag) nil)
      (nil nil)))))
)
 
(remd 'equal) %be quiet when redefined
 
(compiletime (off lapopt))
(lap         %        this is handoptimized version of equal below
   '((*entry EQUAL expr 2)
     (Move (Reg T02) (Reg 1))
     (Move (Reg T03) (Reg 2))
     (Move (Reg S6) -1)
 G0003
     (Move (Reg T01) (Reg S6))
 G0004
     (Xor (Reg S0) (Reg 1) (Reg 2)) %eq test
     (Move (Reg 5) (Reg 1))
     (Rshift (Reg 5) (Reg 5) 59)
     (JSZ G0006)
     (Move (Reg S7) 30)              % id test
     (SUB (Reg S0) (Reg 5) (Reg S7))
     (Move (Reg S6) (Reg 2))
     (Rshift (Reg s6) (Reg s6) 59)
     (JSP nilret)                    % does also NEGint (BOOWOO)
     (Xor (Reg S0) (Reg 5) (Reg S6))
     (Move (Reg 4) (Reg 5))          % R4 is (Tag x)
     (Move (Reg S7) 9)                  %Moved from below
     (JSN nilret)
 
     (Xor (Reg S0) (Reg 4) (Reg S7))  % pair case
     (Move (Reg S6) (Reg T01))
     (Move (Reg S7) 63)
     (JSN G1174)
 
     (Sub (Reg S0) (Reg S6) (Reg S7)) % push
     (Move (Reg A1) (Reg 1))
     (Move (Reg A5) (Reg 2))
     (Load (Reg S6) (Indexed (Reg A1) 1)) %cdr's
     (Load (Reg S7) (Indexed (Reg A5) 1))
     (Load (Reg 1) (Indexed (Reg A1) 0))
     (Load (Reg 2) (Indexed (Reg A5) 0))
     (Move (Reg 5) 1)
     (Move (Reg 4) (Reg T01))
     (Add (Reg 5) (Reg 5) (Reg 4))
     (JSP equalhardcase)   % should not happen very often
     (Move (Reg T01) (Reg 5))
     (Move (Reg A6) (Reg 5))
     (VMoveS (Reg V0) (Reg A6) (Reg S6)) % cdr's are stacked
     (VMoveS (Reg V1) (Reg A6) (Reg S7))
     (J G0004)
 G0006
     (Move (Reg S6) (Reg T01))               %pop
     (Move (Reg 5) -1)
     (Xor (Reg S0) (Reg S6) (Reg 5))
     (Move (Reg A6) (Reg S6))
     (SMoveV (Reg 1) (Reg V0) (Reg A6))
     (SMoveV (Reg 2) (Reg V1) (Reg A6))
     (Move (Reg 4) (Reg T01))
     (Add (Reg S6) (Reg 5) (Reg 4)) %adding while jumping
     (JSN G0003)
     (Move (Reg 1) (Reg T21))
     (Move (Reg 2) 0)               % for gc
     (J (Reg B00))
 Nilret
     (Move (Reg 1) (Reg t00))
     (J (Reg B00))
 G1174
     (Move (Reg S0) (Reg 4)) %posint case
     (JSZ nilret)
 equalhardcase
     (*alloc 6)
     (Move (Reg 2) (Reg T02))
     (Move (Reg 1) (Reg T03))
     (*move (reg t71) (frame 1))
     (*move (reg t72) (frame 2))
     (*move (reg t73) (frame 3))
     (*move (reg t74) (frame 4)) %equal saves t71 - t76 so that
     (*move (reg t75) (frame 5)) %a function calling equal (e.g. assoc)
     (*move (reg t76) (frame 6)) % can be free of alloc in most cases
     (*call ECCI)%
     (*move (frame 1) (reg t71)) % simply put b00, and two more guys
     (*move (frame 2) (reg t72)) % into these regs.
     (*move (frame 3) (reg t73))
     (*move (frame 4) (reg t74))
     (*move (frame 5) (reg t75))
     (*move (frame 6) (reg t76))
     (*exit 6)
))
(compiletime (on  lapopt))
(commentoutcode
(de equal (a b)
     (prog (level x y tgx)
           (setq x a y b)
           (setq level -1)
      eingang
           (cond ((eq x y) (go pop)) % also negint
                 ((or (wgeq (setq tgx (tag x)) id-tag) % bad things!
                      (not (weq tgx (tag y)))) (return nil))
                 ((weq tgx pair-tag) (go push))
                 ((weq tgx posint-tag) (return nil)))
            (return (ecci a b))
 
      push
           (when (wlessp level 63)
                 (setq level (wplus2 level 1))
                 (Vsetelement (vReg 0) level x)
                 (Vsetelement (vReg 1) level y)
                 (setq x (car x))
                 (setq y (car y))
                 (go eingang))
            (return (ecci a b))
 
      pop
            (when (weq level -1) (return t))
            (setq x (cdr (vgetelement (vReg 0) level)))
            (setq y (cdr (vgetelement (vReg 1) level)))
            (setq level (wdifference level 1))
            (go eingang)
))
)
(remd 'atsoc)
 
(compiletime (off lapopt))
 
(if_system CRAY1
 (lap '((*entry atsoc expr 2)
       (*alloc 0)
       (*Move (Reg 2) (Reg a5))
       (*jumpnottype (label returnnil) (Reg 2) pair-tag)
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
      loop
       (*Move 9 (Reg s7))
       (*move (reg 3) (reg s6))
       (Rshift (Reg s6) (Reg s6) 59)
       (Xor (Reg s0) (Reg s6) (Reg s7))
       (*move (reg 3) (reg a5))
       (*Move (Reg 3) (Reg 4))         %saves a Load
       (JSN (label nopair1))
       (*move (indexed (reg a5) 0) (reg 3))
       (*jumpnoteq (label nopair1) (Reg 1) (Reg 3))
       (*Move (Reg 4) (Reg 1))
       (*exit 0)
      nopair1
       (*Move (Reg 5) (Reg 2))
       (*jumpnottype (label returnnil) (Reg 2) pair-tag)
       (*Move (Reg 2) (Reg a5))
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
       (J (label loop))
    returnnil
       (*Move (quote nil) (Reg 1))
       (*exit 0)
)     )
 (lap '((*entry atsoc expr 2)
       (*alloc 0)
       (*Move (Reg 2) (Reg a5))
       (dri)
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
       (*jumpnottype (label returnnil) (Reg 2) pair-tag)
      loop
       (*Move 9 (Reg s7))
       (*Move (Reg 3) (Reg 4))         %saves a Load
       (*move (reg 3) (reg a5))
       (*move (indexed (reg a5) 0) (reg 3))
       (*move (reg 4) (reg s6))
       (Rshift (Reg s6) (Reg s6) 59)
       (Xor (Reg s0) (Reg s6) (Reg s7))
       (JSN (label nopair1))
       (*jumpnoteq (label nopair1) (Reg 1) (Reg 3))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%
       (*Move (Reg 4) (Reg 1))
       (*exit 0)
      nopair1
       (*Move (Reg 5) (Reg 2))
       (*Move (Reg 2) (Reg a5))
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
       (*jumptype (label loop) (Reg 2) pair-tag)
    returnnil
       (*Move (quote nil) (Reg 1))
       (*move (reg 3) (reg 5))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%
       (*exit 0)
))     ) 
 
(if_system CRAY1
 (lap '((*entry atsoc-intern expr 2)
       (*alloc 0)
       (*Move (Reg 2) (Reg a5))
       (*jumpnottype (label returnnil) (Reg 2) pair-tag)
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
      loop
       (*Move 9 (Reg s7))
       (*move (reg 3) (reg s6))
       (Rshift (Reg s6) (Reg s6) 59)
       (Xor (Reg s0) (Reg s6) (Reg s7))
       (*move (reg 3) (reg a5))
       (*Move (Reg 3) (Reg 4))         %saves a Load
       (JSN (label nopair1))
       (*move (indexed (reg a5) 0) (reg 3))
       (*jumpnoteq (label nopair1) (Reg 1) (Reg 3))
       (*Move (Reg 4) (Reg 1))
       (*exit 0)
      nopair1
       (*Move (Reg 5) (Reg 2))
       (*jumpnottype (label returnnil) (Reg 2) pair-tag)
       (*Move (Reg 2) (Reg a5))
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
       (J (label loop))
    returnnil
       (*Move (quote nil) (Reg 1))
       (*exit 0)
)     )
(lap '((*entry atsoc-intern expr 2)
       (*alloc 0)
       (*Move (Reg 2) (Reg a5))
       (dri)
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
       (*jumpnottype (label returnnil) (Reg 2) pair-tag)
      loop
       (*Move 9 (Reg s7))
       (*Move (Reg 3) (Reg 4))         %saves a Load
       (*move (reg 3) (reg a5))
       (*move (indexed (reg a5) 0) (reg 3))
       (*move (reg 4) (reg s6))
       (Rshift (Reg s6) (Reg s6) 59)
       (Xor (Reg s0) (Reg s6) (Reg s7))
       (JSN (label nopair1))
       (*jumpnoteq (label nopair1) (Reg 1) (Reg 3))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%
       (*Move (Reg 4) (Reg 1))
       (*exit 0)
      nopair1
       (*Move (Reg 5) (Reg 2))
       (*Move (Reg 2) (Reg a5))
       (dri) %%%%%%%%%%%%%%%%%%%%%%%
       (Load (Reg 3) (Indexed (Reg a5) 0))
       (Load (Reg 5)  (Indexed (Reg a5) 1))
       (*jumptype (label loop) (Reg 2) pair-tag)
    returnnil
       (*Move (quote nil) (Reg 1))
       (*move (reg s6) (reg 5))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%
       (*exit 0)
))    )
 
(remd 'assoc)
 
(if_system CRAY1
 (lap '((*entry assoc expr 2)
       (*Move (Reg 1) (Reg s7))
       (*move (reg T00) (reg 5))     % second parameter nil?
       (*Wxor3 (Reg s0) (Reg 5) (Reg 2))
       (*wRshift (Reg s7) 59)
       (jsz (label nilret))
 
       (*Move (Reg S7) (reg S0))     % posint case ?
       (*move id-tag (reg s6))       % negint or id case?
       (sub (Reg S6) (Reg S7) (reg S6))
       (jsz (InternalEntry atsoc-intern))
 
       (*Move (Reg S6) (Reg S0))
       (jsP (InternalEntry atsoc-intern))
 
     % nonidcase
       (*move (reg b00) (Reg T75))
       (*Move (Reg 1) (Reg T71))
 loop  (*Move (Reg 2) (Reg T72))
       (*Move (Reg 2) (Reg a5))
       (*WRshift (Reg 2) 59)   %pair test
       (*Move pair-tag (Reg s7))
       (*Wxor3 (Reg s0) (Reg s7) (Reg 2))
       (*Move (Reg t00) (Reg 1))
       (JSN (label ausgang))
       (*Move (Indexed (Reg a5) 0) (Reg S6))
       (*Move (Indexed (Reg a5) 1) (Reg 5))
       (*Move (Reg T71) (Reg 1))   %restore value, have enough time
       (*Move (Reg s6) (Reg T73))
       (*Move (Reg s6) (Reg a2))
       (*wRshift (Reg s6) 59)
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7)) %pair-test
       (*Move (Reg 5) (Reg T74))
       (JSN (label nopair))
       (*Move (Indexed (Reg a2) 0) (Reg 2))
       (*Move (Reg 1) (Reg s6))
       (*wRshift (Reg s6) 59)
       (*Wxor3 (Reg s0) (Reg 1) (Reg 2))  %eq test first
       (*Move (Reg 2) (Reg s7))    % and tags test second
       (*wRshift (Reg s7) 59)
       (jsz (label eqpositiv))
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7))
       (JSN (label nopair))
       (*call equal expr 2)
       (*Move (quote nil) (Reg s7))
       (*Wxor3 (Reg s0) (Reg 1) (Reg s7))
       (*Move (Reg T73) (Reg 1))
       (*move (reg t75) (reg b00))
       (JSN (label ausgang))
  nopair 
       (*Move (Reg T74) (Reg 2))
       (j (label loop))
  eqpositiv
       (*Move (Reg T73) (Reg 1))
  ausgang
       (j (reg b00))
  nilret 
       (*move (reg t00) (Reg 1))
       (j (reg b00))
)    )
 (lap '((*entry assoc expr 2)
       (*Move (Reg 1) (Reg s7))
       (*move (reg T00) (reg 5))     % second parameter nil?
       (*Wxor3 (Reg s0) (Reg 5) (Reg 2))
       (*wRshift (Reg s7) 59)
       (jsz (label nilret))
 
       (*Move (Reg S7) (reg S0))     % posint case ?
       (*move id-tag (reg s6))   % negint or id case?
       (sub (Reg S6) (Reg S7) (reg S6))
       (jsz (InternalEntry atsoc-intern))
 
       (*Move (Reg S6) (Reg S0))
       (jsP (InternalEntry atsoc-intern))
 
     % nonidcase
       (*move (reg b00) (Reg T75))
       (*Move (Reg 1) (Reg T71))
       (dri) %%%%%%%%%%%%
 loop  (*Move (Reg 2) (Reg T72))
       (*Move (Reg 2) (Reg a5))
       (*Move (Indexed (Reg a5) 0) (Reg S6))
       (*Move (Indexed (Reg a5) 1) (Reg 5))
       (*wRshift (Reg 2) 59)   %pair test
       (*Move pair-tag (Reg s7))
       (*Wxor3 (Reg s0) (Reg s7) (Reg 2))
       (*Move (Reg t00) (Reg 1))
       (JSN (label ausgang))
       (*Move (Reg T71) (Reg 1))   %restore value have enough time
       (*Move (Reg s6) (Reg T73))
       (*Move (Reg s6) (Reg a2))
       (*Move (Indexed (Reg a2) 0) (Reg 2))
       (*wRshift (Reg s6) 59)
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7)) %pair-test
       (*Move (Reg 5) (Reg T74))
       (JSN (label nopair))
       (*Move (Reg 1) (Reg s6))
       (*wRshift (Reg s6) 59)
       (*Wxor3 (Reg s0) (Reg 1) (Reg 2))  %eq test first
       (*Move (Reg 2) (Reg s7))    % and tags test second
       (*wRshift (Reg s7) 59)
       (jsz (label eqpositiv))
       (*Wxor3 (Reg s0) (Reg s6) (Reg s7))
       (JSN (label nopair))
       (*call equal expr 2)
       (*Move (quote nil) (Reg s7))
       (*Wxor3 (Reg s0) (Reg 1) (Reg s7))
       (*Move (Reg T73) (Reg 1))
       (*move (reg t75) (reg b00))
       (JSN (label ausgang))
  nopair 
       (*Move (Reg T74) (Reg 2))
       (j (label loop))
  eqpositiv
       (*Move (Reg T73) (Reg 1))
  ausgang
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%%
       (j (reg b00))
  nilret 
       (*move (reg t00) (Reg 1))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%%
       (j (reg b00))
))    )
 
(remd 'member-aux)
 
(if_system CRAY1
 (lap '((*entry member-aux expr 2)
       (*move (reg 2) (reg t72))   %see comment at equal
       (*move (reg 2) (reg a1))
       (*move (indexed (reg a1) 0) (reg 2))
       (*move (reg b00) (reg t73))
  loop
       (*move (reg 1) (reg t71))
       (*call equal)
       (*move (reg t00) (reg s6))
       (*wxor3 (reg s0) (reg s6) (reg 1))
       (*move (reg t72) (reg a1))
       (*move (indexed (reg a1) 1) (reg 2))
       (*move (reg t72) (reg 1))    % result
       (jsn result)
       (*move pair-tag (reg s7))
       (*move (reg t71) (reg 1))
       (*move (reg 2) (reg s6))
       (*wrshift (reg s6) 59)
       (*wxor3 (reg s0) (reg s6) (reg s7))
       (*move (reg 2) (reg t72))
       (*move (reg 2) (reg a1))
       (jsn loopex)
       (*move (indexed (reg a1) 0) (reg 2))
       (j loop)
  loopex 
       (*move (reg t00) (reg 1))  % result nil
       (*move (reg 2) (reg 3))
   result
       (*move (reg t73) (reg b00))
       (j (reg b00))
 ))
 (lap '((*entry member-aux expr 2)
       (*move (reg 2) (reg t72))   %see comment at equal
       (*move (reg 2) (reg a1))
       (*move (indexed (reg a1) 0) (reg 2))
       (dri) %%%%%%%%%
       (*move (reg b00) (reg t73))
  loop
       (*move (reg 1) (reg t71))
       (*call equal)
       (*move (reg t00) (reg s6))
       (*wxor3 (reg s0) (reg s6) (reg 1))
       (*move (reg t72) (reg a1))
       (*move (indexed (reg a1) 1) (reg 2))
       (*move (reg t72) (reg 1))    % result
       (jsn result)
       (*move pair-tag (reg s7))
       (*move (reg t71) (reg 1))
       (*move (reg 2) (reg s6))
       (*wrshift (reg s6) 59)
       (*wxor3 (reg s0) (reg s6) (reg s7))
       (*move (reg 2) (reg t72))
       (*move (reg 2) (reg a1))
       (*move (indexed (reg a1) 0) (reg 2))
       (jsz loop)
 
       (*move (reg t00) (reg 1))  % result nil
       (*move (reg 2) (reg 3))
   result
       (*move (reg t73) (reg b00))
       (eri) %%%%%%%%%%%%%%%%%%%%%%%%%%%
       (j (reg b00))
) ))
 
(if_system CRAY1
 (lap '((*entry !%memq expr 2)
       (*move pair-tag (reg s7))
       (*move (reg 2) (reg s6))
       (*wrshift (reg s6) 59)
       (*wxor3 (reg s0) (reg s6) (reg s7))
       (*move (reg 2) (reg a1))
       (jsn nilret)
  loop
       (*move (indexed (reg a1) 0) (reg 3))
       (*move (indexed (reg a1) 1) (reg 4))
       (*move (Reg 2) (Reg 5))
       (*wxor3 (reg s0) (reg 1) (reg 3))
       (*move (reg 4) (reg 2))
       (*wrshift (reg 4) 59)
       (jsz result)
       (*wxor3 (reg s0) (reg 4) (reg s7))
       (*move (reg 2) (reg a1))
       (jsz loop)
   nilret
       (*move (reg t00) (reg 1))  % result nil
       (j (reg b00))
   result
       (*move (reg 5) (reg 1))    % result
       (j (reg b00))
 ))
  (lap '((*entry !%memq expr 2)
       (*move pair-tag (reg s7))
       (*move (reg 2) (reg s6))
       (*wrshift (reg s6) 59)
       (*wxor3 (reg s0) (reg s6) (reg s7))
       (*move (reg 2) (reg a1))
       (dri) %%%%%%%%
       (*move (indexed (reg a1) 0) (reg 3))
       (*move (indexed (reg a1) 1) (reg 4))
       (jsn nilret)
  loop
       (*move (Reg 2) (Reg 5))
       (*wxor3 (reg s0) (reg 1) (reg 3))
       (*move (reg 4) (reg 2))
       (*move (reg 2) (reg s6))
       (*wrshift (reg s6) 59)
       (jsz result)
       (*wxor3 (reg s0) (reg s6) (reg s7))
       (*move (reg 2) (reg a1))
       (*move (indexed (reg a1) 0) (reg 3))
       (*move (indexed (reg a1) 1) (reg 4))
       (jsz loop)
       (*move (reg 3) (reg 4))
   nilret
       (*move (reg t00) (reg 1))  % result nil
       (eri) %%%%%%%%%%%%
       (j (reg b00))
   result
       (*move (reg 5) (reg 1))    % result
       (eri) %%%%%%%%%%%%
       (j (reg b00))
 )))
 
(compiletime (flag '(hard-put-symfnc1)'internalfunction))
 
(de hard-put-symfnc1()
   (putmem (wplus2 symfnc (id2int 'memq))
                (getmem (wplus2 symfnc (id2int '!%memq)))) nil)
 
(hard-put-symfnc1)
 
(remob '!%memq)
 
(compiletime(on lapopt))
 
(remd 'evlis)
 
(de evlis(u)
      (prog (result pointer)
           (when (not (pairp u)) (return nil))
           (setq pointer (setq result (cons (eval (car u)) nil)))
       loop
           (setq u (cdr u))
           (when (not (pairp u)) (return result))
           (rplacd pointer (cons (eval (car u)) nil))
           (setq pointer (cdr pointer))
           (go loop)))
 
(fluid '(fluid-Reg-vector))
 
%making all fluids which reside in Regs UNBOUND and forming a vector of
%them, which will be used by eval and valuecell
 
(when (pairp fluid-reg-vector)
 (mapcar fluid-Reg-vector (function (lambda (x)
                                      (cond ((null x)    nil)
                                            ((eq x 't)   nil)
                                            (t (makeunbound x))
 )                         )         ) )
 
 (setq fluid-Reg-vector (list2vector fluid-Reg-vector)))
