%
% apollo-sup.sl -- Miscellaneous interface/support routines for Apollo.
%
%       JW Peterson
%       Computer Science Dept.
%       University of Utah
%       12-Feb-83
%
%  Edit by Hucka, 5/25/87.  Stole from 3.2's dumplisp.sl.
%  Edit by Hucka, 4/5/87.  Un-rlisp'ed and converted for 3.4.
%  Edit by Kessler, 04/19/83, 08:09, Added imports statement, to make load
%   of this module autoload everything else that it needs.
%  Edit by JW-Peterson, 7-Jun-83, to squeeze out the holes
%  Edit by JW-Peterson, 16-Apr-84, to add checks to make sure Aegis
%    system routines are where we left them.
%  Edit by JW-Peterson, 6-Jun-84: Increased the amount of "extra" heap
%    allocated from 1K to **16K**!!  Turns out that a LARGE chunk of heap
%    is used between the time the heap is measured and the time it's
%    it's written out.  Previously, this 16K chunk of heap was trashed,
%    causing numerous and random bugs and glitches.
%  Edit by JW-Peterson, 20-Jun-84: Changed CoreSaveEntry to jmp to
%    Apollo_Main instead of Pre-Main, so stack gets reset.
%  Edit by JW-Peterson, 20-jun-86, Fixed so it saves the the total
%    system size. REMAINING BUG: BPS size is still hard-wired.
%
(imports '(all_syscalls syslisp))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Chunk of LAP to move the current image to the mapped one we created.
%
% Call is MoveMemory(MemDest, MemSource, Size);
%

(lap 
        '((*entry movememory expr 3) 
        % Entry with Destination in D0, Source in D1 and Size in D2
        % (move!.l (reg 3) (reg d3))      % use D3 instead of A3 so the flags get set
        (*move (reg 1) (reg a1))
        (*move (reg 2) (reg a2))
movememloop 
       (move!.l (postincrement (reg a2)) (postincrement (reg a1))) 
       (subq!.l 4 (reg 3))             % tick the counter
       (bgt!.s movememloop)             % loop til done
       (rts)))

%
% LAP replacement entry point for a mapped PSL.  Since we want the mapped
% PSL to return directly to the caller (instead of AEGIS) we use a different
% entry/exit mechinism.
%
(lap 
    '((*entry coresaveentry expr 0) 
    (*move (fluid nil) (reg nil))       % set up NIL reg
%   (move!.l 0 (reg a6))                % set up zero A reg
    (*call pre-main) 
    (*exit 0)))

%
% Add a SysCall to create a mapped object.
%
(de ms_crmapl (name                     % FileName
       start                            % Position within object}
       maplen                           % Number of bytes to map}
       conccur)                         % concurreny mode to lock in (0=all read or 
                                    % one writer).
(prog (status address namelen) 
  (setq namelen (stringlength name)) 
  (putarg 1 name '(string)) 
  (putarg 2 namelen '(integer16)) 
  (putarg 3 start '(integer32)) 
  (putarg 4 maplen '(integer32)) 
  (putarg 5 conccur '(integer16)) 
  (cleararg 6) 

  (setq address (syscall 'ms_$crmapl 6)) 
  (setq status (getarg 6 '(integer32) '(fixn))) 
  (cond ((equal status 0) 
         (return address))
        (t (return (apolloerror status))))))

%
% Syscall to unmap PSL once copied.
%

(de ms_unmap (va                        % Virtual address object is mapped at
                  len)                  % Size of object
(prog (status) 
  (putarg 1 va '(integer32)) 
  (putarg 2 len '(integer32)) 
  (cleararg 3) (syscall 
  'ms_$unmap 3) 
  (setq status (getarg 3 '(integer32) '(fixn))) 
  (cond ((equal status 0) (return status)) (t (return (apolloerror status))))))



%
% Procedure to deterimine where a function is in memory.
%
(de addressoffunction (functionname) 
       (return (inf (getfcodepointer functionname))))
             
%
% Read an environment var entry (for snarfing UNMAP_ADDR).  Returns NIL
% if no environment var found, otherwise returns the value as a string.
% We need to resort to SysLisp (YECHH!) because of the string hacking
% needed to handle the pointer to the EV var.
%
(on syslisp)
(de ev_get_var (varname) 
(prog (outlen exists valueptr value) 

  (cond ((idp varname) 
     (setq varname (id2string varname)))) 

  (putarg 1 varname '(string)) 
  (putarg 2 (stringlength varname) '(integer16)) 
  (cleararg 3) 
  (cleararg 4) 
  (setq exists (syscall 'ev_$get_var 4)) 
  (setq value nil) 
  (cond ((weq exists 0)                 % gross recode because un-rlisp choked on original
      (return value))) 

  (setq outlen (getarg 4 '(integer16) '(inum))) 
  (setq valueptr (getarg 3 '(integer32) '(fixn))) 
  (setq value (gtstr (wdifference outlen 1))) (wfor 
  (from i 0 (wdifference outlen 1) 1) (do 
    (putstrbyt value i (getbyte valueptr i)))) 
  (setq value (mkstr value)) 
  (return value)))


